# -*- coding: utf-8 -*-
"""
High-performance raster image operation toolset for SpecPipe

Copyright (c) 2025 Siwei Luo. MIT License.
"""

# Basics
import os

# For local test - delete after use
# Typing
from typing import Callable, Optional, Union, overload

import numpy as np
import rasterio

# GPU acceleration
import torch

# Raster
from rasterio.mask import mask
from rasterio.windows import Window
from shapely.geometry import Polygon
from tqdm import tqdm

# Self
from .specio import simple_type_validator

# %% Crop ROI


# Crop ROI from a raster image and save to a new image
@simple_type_validator
def croproi(
    raster_path: str, roi_coordinates: list[list[tuple[Union[int, float], Union[int, float]]]], output_path: str
) -> None:
    """
    Crop ROI from a raster image and save croped region to a new image.

    Parameters
    ----------
    raster_path : str
        Source raster path.
    roi_coordinates : list[list[tuple[float,float]]]
        Lists of ROI polygon coordinate pairs in tuple.
    output_path : str
        Output raster path.
    """
    # Create a list of Polygons from the coordinate lists
    polygons = [Polygon(poly_coords) for poly_coords in roi_coordinates]

    # Open the raster file
    with rasterio.open(raster_path) as src:
        # Crop the raster
        out_image, out_transform = mask(
            src,
            polygons,
            crop=True,
            all_touched=True,  # Include pixels that touch any polygon
        )

        # Copy the metadata
        out_meta = src.meta.copy()
        out_meta.update(
            {
                "height": out_image.shape[1],
                "width": out_image.shape[2],
                "transform": out_transform,
            }
        )

        # Save the cropped raster
        with rasterio.open(output_path, "w", **out_meta) as dst:
            dst.write(out_image)


# %% Raster tile-wise apply


def dtype_mapper(  # noqa: C901
    dtype: Union[type, str, np.dtype, torch.dtype],
    map_type: str = "numpy",
    min_compatible: bool = True,
) -> Union[type, str, np.dtype, torch.dtype]:
    """
    Map data types between string representations, Python native types, NumPy, PyTorch, and GDAL raster formats.

    Parameters
    ----------
    dtype : Union[type, str]
        Input data type.
    map_type : str, optional
        Target format, can be 'numpy', 'torch' or 'raster'. The default is 'numpy'.
    min_compatible : bool, optional
        Whether minimal compatible data type is returned.
        For example, 'np.uint16' is not supported in PyTorch, then 'torch.int32' is returned.
        The default is True.

    Returns
    -------
    : type
        The mapped data type in the target format.
    """
    torch_map = {
        "float": torch.float32,
        "float16": torch.float16,
        "float32": torch.float32,
        "float64": torch.float64,
        "half": torch.float16,
        "double": torch.float64,
        float: torch.float32,
        np.float16: torch.float16,
        np.float32: torch.float32,
        np.float64: torch.float64,
        "int": torch.int32,
        "int8": torch.int8,
        "int16": torch.int16,
        "int32": torch.int32,
        "int64": torch.int64,
        int: torch.int32,
        np.int8: torch.int8,
        np.int16: torch.int16,
        np.int32: torch.int32,
        np.int64: torch.int64,
        "uint8": torch.uint8,
        np.uint8: torch.uint8,
        "bool": torch.bool,
        bool: torch.bool,
        np.bool_: torch.bool,
    }
    torch_ext = {
        "uint16": torch.int32,
        "uint32": torch.int32,
        np.uint16: torch.int32,
        np.uint32: torch.int32,
    }
    np_map = {
        "float": np.float32,
        "float16": np.float16,
        "float32": np.float32,
        "float64": np.float64,
        "half": np.float16,
        "double": np.float64,
        float: np.float32,
        torch.float16: np.float16,
        torch.float32: np.float32,
        torch.float64: np.float64,
        "int": np.int32,
        "int8": np.int8,
        "int16": np.int16,
        "int32": np.int32,
        "int64": np.int64,
        int: np.int32,
        torch.int8: np.int8,
        torch.int16: np.int16,
        torch.int32: np.int32,
        torch.int64: np.int64,
        "uint8": np.uint8,
        torch.uint8: np.uint8,
        "uint16": np.uint16,
        "uint32": np.uint32,
        "uint64": np.uint64,
        "bool": np.bool_,
        bool: np.bool_,
        torch.bool: np.bool_,
    }
    raster_type = {
        "float": "float32",
        "double": "float32",
        float: "float32",
        torch.float32: "float32",
        torch.float64: "float64",
        np.float32: "float32",
        np.float64: "float64",
        int: "int32",
        torch.int16: "int16",
        torch.int32: "int32",
        np.int16: "int16",
        np.int32: "int32",
        torch.uint8: "uint8",
        np.uint8: "uint8",
        np.uint16: "uint16",
        bool: "bool",
    }
    raster_ext = {
        "float16": "float32",
        torch.float16: "float32",
        np.float16: "float32",
        "half": "float32",
        "int8": "int16",
        np.int8: "int16",
        torch.int8: "int16",
        "uint32": "int32",
        np.uint32: "int32",
    }
    if map_type.lower() == "numpy":
        if dtype in list(np_map.values()):
            result_np: Union[type, str, np.dtype, torch.dtype] = dtype
            return dtype
        elif dtype in list(np_map.keys()):
            result_np = np_map[dtype]
            return result_np
        else:
            raise ValueError(f"Unsupported NumPy data type: {dtype}")
    elif map_type.lower() == "torch":
        if dtype in list(torch_map.values()):
            result_torch: Union[type, str, np.dtype, torch.dtype] = dtype
            return result_torch
        elif dtype in list(torch_map.keys()):
            result_torch = torch_map[dtype]
            return result_torch
        elif min_compatible & (dtype in list(torch_ext.keys())):
            return torch_ext[dtype]
        else:
            raise ValueError(f"Unsupported PyTorch data type: {dtype}")
    elif map_type.lower() == "raster":
        if dtype in list(raster_type.values()):
            result_ras: Union[type, str, np.dtype, torch.dtype] = dtype
            return result_ras
        elif dtype in list(raster_type.keys()):
            result_ras = raster_type[dtype]
            return result_ras
        elif min_compatible & (dtype in list(raster_ext.keys())):
            result_ras = raster_ext[dtype]
            return result_ras
        else:
            raise ValueError(f"Unsupported raster data type: {dtype}")
    else:
        raise ValueError(f"Unsupported map_type, map_type must be 'numpy', 'torch' or 'raster', got: '{map_type}'")


@overload
def auto_fp(
    data: torch.Tensor,
    scaling: bool = False,
    rtol: float = 1e-4,
    atol: float = 1e-5,
    safety_factor: float = 2.0,
) -> tuple[torch.Tensor, Optional[float]]: ...


@overload
def auto_fp(  # type: ignore[overload-cannot-match]
    data: np.ndarray,
    scaling: bool = False,
    rtol: float = 1e-4,
    atol: float = 1e-5,
    safety_factor: float = 2.0,
) -> tuple[np.ndarray, Optional[float]]: ...


# Mypy failure on GitHub, the code works and passes local mypy validation


def auto_fp(  # noqa: C901
    data: Union[np.ndarray, torch.Tensor],
    scaling: bool = False,
    rtol: float = 1e-4,
    atol: float = 1e-5,
    safety_factor: float = 2.0,
) -> tuple[Union[np.ndarray, torch.Tensor], Optional[float]]:
    """
    Cast numeric data in PyTorch tensor to type FP16 if compatible.
    If scaling set True, the function will try to scale the data to fit FP16, and scaling coefficient is returned if compatible.

    Parameters
    ----------
    data : Union[np.ndarray, torch.Tensor]
        Input data in NumPy ndarray or PyTorch tensor.

    scaling : bool, optional
        If data is scaled to fit FP16. The default is False.

    rtol : float, optional
        Relative tolerance for precision checks. The default is 1e-4.

    atol : float, optional
        Absolute tolerance for small-value checks. The default is 1e-5.

    safety_factor : float, optional
        Multiplier to avoid FP16 limits. The default is 2.0.

    Returns
    -------
    tuple[torch.Tensor, Optional[float]]
        - Scaled data in FP16, if FP16 is compatible after scaling, and scaling factor.
        - Data in FP16, if FP16 is natively compatible, and 1.0.
        - Original data, if FP16 is not compatible or not compatible despite scaling if scaling set True.
    """  # noqa: E501
    # FP16 limits
    fp16_max = 65504 / safety_factor
    fp16_min = 6.10e-5 * safety_factor

    ## For numpy array
    if type(data) is np.ndarray:
        short_types = [np.int8, np.uint8, np.int16, np.uint16, np.float16]
        long_types = [np.int32, np.int64, np.uint32, np.uint64, np.float32, np.float64]

        if data.dtype in short_types:
            return (data, None)

        elif data.dtype in long_types:
            # Get min and max values of non_zeros
            max_val: float = float(np.max(np.abs(data)))
            non_zero_data = data[data != 0]
            if non_zero_data.size > 0:
                min_val: float = float(np.min(np.abs(non_zero_data)))
            else:
                min_val = fp16_min

            # If data fits in FP16 natively
            if max_val <= fp16_max and min_val >= fp16_min:
                arr_fp16: np.ndarray = data.astype(np.float16)
                if np.allclose(data, arr_fp16.astype(data.dtype), rtol=rtol, atol=atol):
                    return (arr_fp16, 1.0)

            # If data can be scaled into FP16 range
            if scaling:
                scale_coef = 1.0
                # Scale down if too large
                if max_val > fp16_max:
                    scale_coef = fp16_max / max_val
                # Scale up if too small
                elif min_val < fp16_min:
                    scale_coef = fp16_min / min_val
                # Scale up if value is in the FP16 range but too small to meet precision requirements in conversion
                elif (
                    np.allclose(data, arr_fp16.astype(data.dtype), rtol=rtol, atol=atol) is False
                    and (fp16_max * (1 / min_val)) < fp16_max
                ):
                    scale_coef = 1 / min_val

                # Check if scaled data preserves precision
                scaled_arr = data * scale_coef
                scaled_fp16_arr = scaled_arr.astype(np.float16)
                # Compare loss with tolerance threshold
                if np.allclose(
                    scaled_arr.astype(scaled_arr.dtype),
                    scaled_fp16_arr.astype(scaled_arr.dtype),
                    rtol=rtol,
                    atol=atol,
                ):
                    return (scaled_fp16_arr, scale_coef)

            # If FP16 is incompatible
            return (data, None)  # FP32 required, e.g. mixed large / small values

        else:
            raise ValueError(f"Unsupported data dtype '{data.dtype}'")

    ## For torch tensor
    elif type(data) is torch.Tensor:
        if data.dtype is torch.float16:
            return (data, None)

        elif (data.dtype is torch.float32) or (data.dtype is torch.float64):
            # Get min and max values of non_zeros
            max_val = float(data.abs().max())
            non_zero_data = data[data != 0]
            if non_zero_data.numel() > 0:
                min_val = float(non_zero_data.abs().min())
            else:
                min_val = fp16_min

            # If data fits in FP16 natively
            if max_val <= fp16_max and min_val >= fp16_min:
                torch_fp16 = data.to(torch.float16)
                if torch.allclose(data, torch_fp16.to(data.dtype), rtol=rtol, atol=atol):
                    return (torch_fp16, 1.0)

            # If data can be scaled into FP16 range or FP16 precision
            if scaling:
                # Scale down if too large
                if max_val > fp16_max:
                    scale_coef = fp16_max / max_val
                # Scale up if too small
                elif min_val < fp16_min:
                    scale_coef = fp16_min / min_val
                # Scale up if value is in the FP16 range but too small to meet precision requirements in conversion
                elif (
                    torch.allclose(data, torch_fp16.to(data.dtype), rtol=rtol, atol=atol) is False
                    and (fp16_max * (1 / min_val)) < fp16_max
                ):
                    scale_coef = 1 / min_val

                # Check if scaled data preserves precision
                scaled_tensor = data * scale_coef
                scaled_fp16_tensor = scaled_tensor.to(torch.float16)
                if torch.allclose(
                    scaled_tensor.to(scaled_tensor.dtype),
                    scaled_fp16_tensor.to(scaled_tensor.dtype),
                    rtol=rtol,
                    atol=atol,
                ):
                    return (scaled_fp16_tensor, scale_coef)

            # If FP16 is incompatible
            return (data, None)  # FP32 required, e.g. mixed large / small values

        else:
            raise ValueError(f"Unsupported data dtype '{data.dtype}'")

    else:
        raise ValueError(f"Unsupported data type '{type(data)}'")


# Tiled pixel-wise apply a 1D function
@simple_type_validator
def pixel_spec_apply(
    image_path: str,
    output_path: str,
    spectral_function: Callable,
    dtype: Union[type, str, np.dtype, torch.dtype] = "float32",
    tile_size: int = -1,
    progress: bool = True,
) -> None:
    """
    Apply a function to the 1D spectra of every pixel of a raster image.
    The function must accept 1D arraylike as only required parameter, and return processed 1D arraylike data.
    """
    # Validate tile size and set default
    if tile_size == -1:
        tile_size = 32
    elif tile_size < 1:
        raise ValueError(f"If provided, tile_size must be positive, got: {tile_size}")

    # Validate dtype
    dtype = dtype_mapper(dtype)

    with rasterio.open(image_path) as src:
        # Validate dst raster
        if os.path.exists(output_path):
            os.remove(output_path)

        # Get output number of bands
        test_data = src.read(window=Window(col_off=0, row_off=0, width=1, height=1))
        test_result = spectral_function(test_data[:, 0, 0])
        num_bands_out = len(test_result)

        # Get metadata from the source image
        meta = src.meta.copy()
        meta.update({"dtype": dtype_mapper(dtype, "raster")})
        meta.update({"count": num_bands_out})

        # Create dst raster
        with rasterio.open(output_path, "w", **meta) as dst:
            # Get image dimensions
            height, width = src.height, src.width

            # Process in tiles
            if progress:
                print(f"\nProcessing image with tile size {tile_size}x{tile_size}\n")
            for i in tqdm(range(0, height, tile_size), total=int(height / tile_size) + 1, disable=not progress):
                for j in range(0, width, tile_size):
                    # Define window for current tile
                    win = Window(
                        col_off=j,
                        row_off=i,
                        width=min(tile_size, width - j),
                        height=min(tile_size, height - i),
                    )
                    # Read all bands, shape: [bands, rows, cols]
                    tile_data = src.read(window=win)
                    tile_height, tile_width = tile_data.shape[1], tile_data.shape[2]
                    # Create array and save processed data
                    processed_tile = np.zeros((num_bands_out, tile_height, tile_width))
                    for m in range(tile_height):
                        for n in range(tile_width):
                            # Apply function
                            processed_spec = spectral_function(tile_data[:, m, n])
                            processed_tile[:, m, n] = np.array(processed_spec)
                    # Write array to output
                    dst.write(processed_tile, window=win)


# Tiled vectorized apply a 2D function
@simple_type_validator
def pixel_array_apply(
    image_path: str,
    output_path: str,
    spectral_function: Callable,
    dtype: Union[type, str, np.dtype, torch.dtype] = "float32",
    tile_size: int = -1,
    progress: bool = True,
) -> None:
    """
    Apply a function to the 1D spectra of every pixel of a raster image.
    The function must accept 2D array as only required parameter, and return processed 2D array with same length.
    Each row of the 2D array represents an 1D spectra or processed data of a pixel
    """
    # Validate tile size and set default
    if tile_size == -1:
        tile_size = 32
    elif tile_size < 1:
        raise ValueError(f"If provided, tile_size must be positive, got: {tile_size}")

    # Validate dtype
    dtype = dtype_mapper(dtype)

    with rasterio.open(image_path) as src:
        # Validate dst raster
        if os.path.exists(output_path):
            os.remove(output_path)

        # Get output number of bands
        test_data = src.read(window=Window(col_off=0, row_off=0, width=2, height=2))
        # Reshape test data
        test_data_input = test_data.reshape(src.count, -1).T
        test_result = spectral_function(test_data_input)
        num_bands_out = np.array(test_result).shape[1]

        # Get metadata from the source image
        meta = src.meta.copy()
        meta.update({"dtype": dtype_mapper(dtype, "raster")})
        meta.update({"count": num_bands_out})

        # Create dst raster
        with rasterio.open(output_path, "w", **meta) as dst:
            # Get image dimensions
            height, width = src.height, src.width

            # Process in tiles
            if progress:
                print(f"\nProcessing image with tile size {tile_size}x{tile_size}\n")
            for i in tqdm(range(0, height, tile_size), total=int(height / tile_size) + 1, disable=not progress):
                for j in range(0, width, tile_size):
                    # Define window for current tile
                    win = Window(
                        col_off=j,
                        row_off=i,
                        width=min(tile_size, width - j),
                        height=min(tile_size, height - i),
                    )
                    # Read all bands for current tile, shape: [bands, rows, cols]
                    tile_data = src.read(window=win)
                    original_shape = tile_data.shape
                    # Reshape src tile data
                    spectra_2d = tile_data.reshape(src.count, -1).T
                    spectra_2d = np.array(spectra_2d)
                    # Apply vectorized function
                    processed_spectra = spectral_function(spectra_2d)
                    # Reshape back
                    output_shape = (num_bands_out, original_shape[1], original_shape[2])
                    processed_tile = processed_spectra.T.reshape(output_shape)
                    # Write processed tile to output
                    dst.write(processed_tile, window=win)


# Tiled tensor apply a tensor function with calculation axis = 0
@simple_type_validator
def pixel_tensor_apply(  # noqa: C901
    image_path: str,
    output_path: str,
    spectral_function: Callable,
    dtype: Union[type, str, np.dtype, torch.dtype] = "float32",
    tile_size: int = -1,
    device: str = "cuda",
    progress: bool = True,
) -> None:
    """
    Apply a function to the 1D spectra of every pixel of a raster image.
    The function must accept 3D PyTorch tenser as only required parameter, and return processed 3D tensor with same size.
    The function must compute along 0 axis. And the input and output tensor must have the same size except 0 axis.
    """  # noqa: E501
    # Validate tile size and set default
    if tile_size == -1:
        tile_size = 64
    elif tile_size < 1:
        raise ValueError(f"If provided, tile_size must be positive, got: {tile_size}")

    # Validate dtype
    dtype = dtype_mapper(dtype, "torch")

    # Initialize torch device
    device = str(torch.device(device if torch.cuda.is_available() and device == "cuda" else "cpu"))
    if progress:
        print(f"\nUsing device: {device}")
        if "cuda" in str(device):
            print(f"GPU: {torch.cuda.get_device_name(0)}")

    # Use CUDA Stream
    stream = torch.cuda.Stream() if "cuda" in str(device) else None

    with rasterio.open(image_path) as src:
        # Validate dst raster for cuda output
        if os.path.exists(output_path):
            os.remove(output_path)

        # Data for test running of given function
        test_data = src.read(window=Window(col_off=0, row_off=0, width=2, height=2))

        # Validate data type
        nctypes = [np.uint16, np.uint32, np.uint64]
        if test_data.dtype in nctypes:
            test_data = test_data.astype(dtype_mapper(dtype))
        test_tensor = torch.from_numpy(test_data).to(device)

        # Get output number of bands of given function
        test_tensor_out = spectral_function(test_tensor)
        num_bands_out = test_tensor_out.shape[0]
        del test_tensor_out
        if torch.cuda.is_available():
            torch.cuda.empty_cache()

        # Get metadata from the source image
        meta = src.meta.copy()
        meta.update({"dtype": dtype_mapper(dtype, "raster")})
        meta.update({"count": num_bands_out})

        # Create output file with same number of bands
        with rasterio.open(output_path, "w", **meta) as dst:
            # Get image dimensions
            height, width = src.height, src.width

            # Process in tiles
            if progress:
                print(f"\nProcessing image with tile size {tile_size}x{tile_size}\n")

            for i in tqdm(range(0, height, tile_size), total=int(height / tile_size) + 1, disable=not progress):
                for j in range(0, width, tile_size):
                    # Define window for current tile
                    win = Window(
                        col_off=j,
                        row_off=i,
                        width=min(tile_size, width - j),
                        height=min(tile_size, height - i),
                    )

                    # Read all bands for current tile, shape: [bands, rows, cols]
                    tile_data = src.read(window=win)

                    # Validate data type
                    if tile_data.dtype in nctypes:
                        tile_data = tile_data.astype(dtype_mapper(dtype))

                    # Create tensor from tile-data and move to GPU
                    with torch.cuda.stream(stream):
                        tile_tensor = torch.from_numpy(tile_data).pin_memory().to(device, non_blocking=True)

                    # Convert tensor dtype
                    if tile_tensor.dtype is not dtype_mapper(dtype, "torch"):
                        assert isinstance(dtype, (str, torch.dtype))
                        tile_tensor = tile_tensor.to(dtype)

                    # Apply tensor function
                    with (
                        torch.cuda.stream(stream),
                        torch.no_grad(),
                    ):  # Disable gradient tracking for memory efficiency
                        processed = spectral_function(tile_tensor)

                    # Move back to CPU and convert to numpy
                    with torch.cuda.stream(stream):
                        processed = processed.cpu().numpy()

                    # Synchronize before writing
                    if stream:
                        torch.cuda.synchronize()

                    # Write output
                    dst.write(processed, window=win)


# Row-wise tensor apply a tensor function with calculation axis = 1
@simple_type_validator
def pixel_tensor_hyper_apply(  # noqa: C901
    image_path: str,
    output_path: str,
    spectral_function: Callable,
    dtype: Union[type, str, np.dtype, torch.dtype] = "float32",
    tile_size: int = -1,
    device: str = "cuda",
    progress: bool = True,
) -> None:
    """
    Apply a function to the 1D spectra of every pixel of a raster image, optimized for hyperspectral image data transfer.
    The function must accept 3D PyTorch tenser as only required parameter, and return processed 3D tensor with same size.
    The function must compute along 1 axis. And the input and output tensor must have the same size except 1 axis.
    """  # noqa: E501
    # Validate tile size and set default
    if tile_size == -1:
        row_chunk = 4
    elif tile_size < 1:
        raise ValueError(f"Row_chunk must be positive, got: {tile_size}")
    else:
        row_chunk = tile_size

    # Validate dtype
    dtype = dtype_mapper(dtype, "torch")

    # Initialize torch device
    device = str(torch.device(device if torch.cuda.is_available() and device == "cuda" else "cpu"))
    if progress:
        print(f"\nUsing device: {device}")
        if "cuda" in str(device):
            print(f"GPU: {torch.cuda.get_device_name(0)}")

    # Use CUDA Stream
    stream = torch.cuda.Stream() if "cuda" in str(device) else None

    # Validate dst raster for cuda output
    if os.path.exists(output_path):
        os.remove(output_path)

    with rasterio.open(image_path) as src:
        # Data for test running of given function
        test_data = src.read(window=Window(0, 0, src.width, row_chunk))

        # Validate data type
        nctypes = [np.uint16, np.uint32, np.uint64]
        if test_data.dtype in nctypes:
            test_data = test_data.astype(dtype_mapper(dtype))

        # Get output number of bands of given function
        test_tensor = torch.from_numpy(test_data).to(device)  # [C, 1, W]
        test_tensor = test_tensor.permute(1, 0, 2)  # [1, C, W]
        with torch.no_grad():
            test_out = spectral_function(test_tensor)
        num_bands_out = test_out.shape[1]
        del test_data, test_tensor, test_out
        if torch.cuda.is_available():
            torch.cuda.empty_cache()

        # Setup output
        meta = src.meta.copy()
        meta.update({"dtype": dtype_mapper(dtype, "raster"), "count": num_bands_out})

        with rasterio.open(output_path, "w", **meta) as dst:
            for row_start in tqdm(range(0, src.height, row_chunk), disable=not progress):
                # Read entire row chunk [C, row_chunk, W]
                window = Window(0, row_start, src.width, min(row_chunk, src.height - row_start))
                row_data = src.read(window=window)

                # Validate data type
                if row_data.dtype in nctypes:
                    row_data = row_data.astype(dtype_mapper(dtype))

                # Convert to [row_chunk, C, W] for processing
                with torch.cuda.stream(stream):
                    row_tensor = torch.from_numpy(row_data).pin_memory().to(device, non_blocking=True)
                row_tensor = row_tensor.permute(1, 0, 2)  # [row_chunk, C, W]

                # Convert tensor dtype
                if row_tensor.dtype is not dtype_mapper(dtype, "torch"):
                    assert isinstance(dtype, (str, torch.dtype))
                    row_tensor = row_tensor.to(dtype)

                # Apply tensor function
                with torch.cuda.stream(stream), torch.no_grad():
                    processed = spectral_function(row_tensor)

                # Convert back to [C, row_chunk, W] for writing
                with torch.cuda.stream(stream):
                    processed_cpu = processed.permute(1, 0, 2).cpu().numpy()

                # Synchronize before writing
                if stream:
                    torch.cuda.synchronize()

                # Write output
                dst.write(processed_cpu, window=window)


# Raster pixel apply - apply a function to every raster pixel
@simple_type_validator
def pixel_apply(
    image_path: str,
    spectral_function: Callable,
    function_type: str,
    output_path: Optional[str] = None,
    dtype: Union[type, str, np.dtype, torch.dtype] = "float32",
    tile_size: int = -1,
    progress: bool = True,
    return_output_path: bool = True,
) -> Optional[str]:
    """
    Apply a function to process the 1D spectra of every pixel of a raster image.

    Parameters
    ----------
    image_path : str
        Input raster image path in string.

    output_path : str
        Output raster image path in string.

    spectral_function : Callable, optional
        Function to apply to 1D spectra of every pixel.
        The type of the function is specified to parameter function_type

    function_type : Callable, optional
        Specifies the type of spectral processing function to apply. Must be one of:
        - 'spec' :
            Processes individual spectra. The function must:
            - Accept a 1D array-like as its only required input.
            - Return 1D array of the processed spectrum.
        - 'array' :
            Processes batches of spectra as a 2D array. The function must:
            - Accept a 2D array as its only required input, where each row represents a spectrum of a pixel.
            - Return a 2D array with the same number of rows (processed spectra).
        - 'tensor' :
            Processes data as a 3D PyTorch tensor, optimized for multispectral data with limited number of bands. 'cuda' is applied if available.
            The function must:
            - Accept a 3D tensor as its only required input (shape: [Channels, Height, Width]).
            - Return a 3D tensor with identical shape except Channel axis (axis 0).
            - Compute operations along the Channel dimension (axis 0).
        - 'tensor_hyper' :
            Processes data as a 3D PyTorch tensor, optimized for hyperspectral data with large number of bands. 'cuda' is applied if available.
            The function requirements are identical with the 'tensor' type, except:
            - Compute operations along axis 1.

    dtype : Union[type, str], optional
        Value data type of output array. The default is 'float32'.

    tile_size : int, optional
        Size of the processing tiles in pixels.
        Larger values may improve performance, but require more memory.
        Default for function type 'spec' and 'array' is 32, for 'tensor' is 64, and for

    progress : bool, optional
        If True, enables progress bar. If False, suppresses progress messages. Default is True.

    return_output_path : bool, optional
        Whether path of processed image is returned. Default is True.

    Returns
    -------
    output_path : str
        Path of processed image.
    """  # noqa: E501
    # Default output path
    if output_path is None:
        ptx = os.path.splitext(image_path)
        output_path = ptx[0] + "_px_app_" + spectral_function.__name__ + ptx[1]

    # Validate path
    if not os.path.exists(image_path):
        raise ValueError(f"\nInvalid image_path: {image_path} \nImage file does not exist.")
    output_dir_path = os.path.dirname(output_path)
    if not os.path.exists(output_dir_path):
        raise ValueError(f"\nInvalid output directory: {output_dir_path} \nDirectory does not exist.")

    # Validate tile size and batch size
    if (tile_size != -1) & (tile_size < 1):
        raise ValueError(f"\nInvalid tile_size, got: {tile_size} \nIf provided, tile_size must be positive integer.")

    # Apply function
    # Spec apply
    if function_type.lower() == "spec":
        pixel_spec_apply(image_path, output_path, spectral_function, dtype, tile_size, progress)

    # Array apply
    elif function_type.lower() == "array":
        pixel_array_apply(image_path, output_path, spectral_function, dtype, tile_size, progress)

    # Tensor apply
    elif function_type.lower() == "tensor":
        pixel_tensor_apply(image_path, output_path, spectral_function, dtype, tile_size, "cuda", progress)

    # Hyper-tensor apply
    elif function_type.lower() == "tensor_hyper":
        pixel_tensor_hyper_apply(image_path, output_path, spectral_function, dtype, tile_size, "cuda", progress)

    # Else
    else:
        raise ValueError(
            f"Invalid function_type, function_type must be 'spec', 'array' or 'tensor', but got: {function_type}"
        )

    # Return path of processed image
    if return_output_path:
        return output_path
    else:
        return None



# -*- coding: utf-8 -*-
"""
Tests for modeling and model evaluation module (ModelEva)

Copyright (c) 2025 Siwei Luo. MIT License.
"""

# Test
# OS Files
import os
import shutil

# For local test - delete after use
import tempfile
import unittest

# Testing third
import numpy as np
import pandas as pd
import pytest

# Modeling
from sklearn.linear_model import LinearRegression, LogisticRegression

# Self
# Funcs to test
from specpipe.modeleva import ModelEva

# %% Helper functions for ModelEva testing


def create_sample_data_regression(
    n_samples: int = 10,
) -> list[tuple[str, tuple[int], float, np.ndarray]]:
    """Create regression sample data."""
    np.random.seed(66)
    X = np.random.rand(n_samples, 5)  # noqa: N806
    y = np.random.rand(n_samples)
    return [(f"sample_{i}", (5,), float(y[i]), X[i]) for i in range(n_samples)]


def create_sample_data_classification(
    n_samples: int = 10,
) -> list[tuple[str, tuple[int], str, np.ndarray]]:
    """Create classification sample data."""
    np.random.seed(66)
    X = np.random.rand(n_samples, 5)  # noqa: N806
    y = np.random.choice(["A", "B"], size=n_samples)
    return [(f"sample_{i}", (5,), str(y[i]), X[i]) for i in range(n_samples)]


def modeleva_initialization_regression(temp_dir: str, data_split: str = "5-fold", n_samples: int = 10) -> ModelEva:
    """Test regression model initialization."""
    sample_list = create_sample_data_regression(n_samples)

    model = LinearRegression()

    report_dir = str(temp_dir).replace("\\", "/")
    if not os.path.exists(report_dir):
        os.makedirs(report_dir)

    model_eva = ModelEva(
        sample_list=sample_list,
        model=model,
        validation_method=data_split,
        report_directory=report_dir,
        data_label="test_reg_data",
        is_regression=True,
        silent_all=True,
    )

    return model_eva


def modeleva_initialization_classification(temp_dir: str, data_split: str = "5-fold", n_samples: int = 10) -> ModelEva:
    """Test classification model initialization."""
    sample_list = create_sample_data_classification(n_samples)

    model = LogisticRegression()

    report_dir = str(temp_dir).replace("\\", "/")
    if not os.path.exists(report_dir):
        os.makedirs(report_dir)

    model_eva = ModelEva(
        sample_list=sample_list,
        model=model,
        validation_method=data_split,
        report_directory=report_dir,
        data_label="test_cls_data",
        is_regression=False,
        silent_all=True,
    )

    return model_eva


# %% test functions : ModelEva


class TestModelEva:
    """Test class for ModelEva functionality."""

    @staticmethod
    def test_initialization_regression() -> None:
        """Test regression model initialization."""
        with tempfile.TemporaryDirectory() as temp_dir:
            model_eva = modeleva_initialization_regression(temp_dir)

            assert os.path.exists(model_eva.report_directory)
            assert type(model_eva.model) is LinearRegression
            assert model_eva.model_label == LinearRegression.__name__
            assert model_eva.is_regression is True
            assert model_eva.X.shape == (10, 5)
            assert model_eva.y.shape == (10, 1)
            assert model_eva.validation_method == 5

        # Clear test report dir
        if os.path.exists(temp_dir):
            shutil.rmtree(temp_dir)

    @staticmethod
    def test_initialization_classification() -> None:
        """Test classification model initialization."""
        with tempfile.TemporaryDirectory() as temp_dir:
            model_eva = modeleva_initialization_classification(temp_dir)

            assert os.path.exists(model_eva.report_directory)
            assert type(model_eva.model) is LogisticRegression
            assert model_eva.model_label == LogisticRegression.__name__
            assert model_eva.is_regression is False
            assert model_eva.X.shape == (10, 5)
            assert model_eva.y.shape == (10, 1)
            assert model_eva.validation_method == 5

        # Clear test report dir
        if os.path.exists(temp_dir):
            shutil.rmtree(temp_dir)

    @staticmethod
    def test_auto_choose_classification_regression() -> None:
        """Test automatic determination of regression and classification using target type."""
        with tempfile.TemporaryDirectory() as temp_dir:
            # Create report direcotry
            report_dir = str(temp_dir).replace("\\", "/")
            if not os.path.exists(report_dir):
                os.makedirs(report_dir)

            # Auto regression
            sample_list_reg = create_sample_data_regression(10)
            model = LinearRegression()
            model_eva = ModelEva(
                sample_list=sample_list_reg,
                model=model,
                validation_method="10-fold",
                report_directory=report_dir,
                data_label="test_reg_data",
                silent_all=True,
            )
            assert model_eva.is_regression is True

            # Auto classification
            sample_list_cls = create_sample_data_classification(10)
            model = LogisticRegression()
            model_eva = ModelEva(
                sample_list=sample_list_cls,
                model=model,
                validation_method="10-fold",
                report_directory=report_dir,
                data_label="test_reg_data",
                silent_all=True,
            )
            assert model_eva.is_regression is False

        # Clear test report dir
        if os.path.exists(temp_dir):
            shutil.rmtree(temp_dir)

    @staticmethod
    def test_auto_kfold_to_loocv() -> None:
        """Test automatic convert kfold to loo if k larger than sample size."""
        with tempfile.TemporaryDirectory() as temp_dir:
            # Create report direcotry
            report_dir = str(temp_dir).replace("\\", "/")
            if not os.path.exists(report_dir):
                os.makedirs(report_dir)

            # Create model evaluator
            sample_list = create_sample_data_regression(8)
            model = LinearRegression()
            with pytest.warns(UserWarning, match="'loo' is applied instead"):
                model_eva = ModelEva(
                    sample_list=sample_list,
                    model=model,
                    validation_method="10-fold",
                    report_directory=report_dir,
                    data_label="test_reg_data",
                    is_regression=True,
                    silent_all=True,
                )
            assert len(model_eva.sid) == 8
            assert model_eva.validation_method == "loo"

        # Clear test report dir
        if os.path.exists(temp_dir):
            shutil.rmtree(temp_dir)

    @staticmethod
    def test_property_access() -> None:
        """Test read-only property access."""
        with tempfile.TemporaryDirectory() as temp_dir:
            model_eva = modeleva_initialization_classification(temp_dir)
            with pytest.raises(ValueError, match="cannot be modified"):
                model_eva.create_time = "2025-01-01"
            with pytest.raises(ValueError, match="cannot be modified"):
                model_eva.model_time = "2025-01-01"
            with pytest.raises(ValueError, match="cannot be modified"):
                model_eva.unseen_threshold = 0.0
            with pytest.raises(ValueError, match="cannot be modified"):
                model_eva.sid = np.array([["1", "2", "3"]])
            with pytest.raises(ValueError, match="cannot be modified"):
                model_eva.X_original_shape = (2, 2)
            with pytest.raises(ValueError, match="cannot be modified"):
                model_eva.y = [1, 2, 3]
            with pytest.raises(ValueError, match="cannot be modified"):
                model_eva.X = [[0, 1], [1, 2], [1, 2]]
            with pytest.raises(ValueError, match="cannot be modified"):
                model_eva.dsp_inds = [("1", "2"), ("2", "1")]
            with pytest.raises(ValueError, match="cannot be modified"):
                model_eva.y_true_eva = [1, 2, 3]
            with pytest.raises(ValueError, match="cannot be modified"):
                model_eva.y_pred_eva = [1, 2, 3]
            with pytest.raises(ValueError, match="cannot be modified"):
                model_eva.ynames = ["1", "2", "3"]
            with pytest.raises(ValueError, match="cannot be modified"):
                model_eva.y_true_proba_eva = [[0.1, 0.9], [0.1, 0.9], [0.1, 0.9]]
            with pytest.raises(ValueError, match="cannot be modified"):
                model_eva.y_pred_proba_eva = [[0.1, 0.9], [0.1, 0.9], [0.1, 0.9]]
            with pytest.raises(ValueError, match="cannot be modified"):
                model_eva.sid_eva = ["1", "2", "3"]
            with pytest.raises(ValueError, match="cannot be modified"):
                model_eva.app_model = LinearRegression()
            with pytest.raises(ValueError, match="cannot be modified"):
                model_eva.app_model_create_time = "2025-01-01"

        # Clear test report dir
        if os.path.exists(temp_dir):
            shutil.rmtree(temp_dir)

    @staticmethod
    def test_update_samples() -> None:
        """Test sample update functionality."""
        with tempfile.TemporaryDirectory() as temp_dir:
            # Classification
            model_eva = modeleva_initialization_classification(temp_dir)
            new_samples_cls = create_sample_data_classification(11)
            model_eva.update_samples(new_samples_cls)

            assert model_eva.X.shape == (11, 5)
            assert model_eva.y.shape == (11, 1)

            # Regression
            model_eva = modeleva_initialization_regression(temp_dir)
            new_samples_reg = create_sample_data_regression(12)
            model_eva.update_samples(new_samples_reg)

            assert model_eva.X.shape == (12, 5)
            assert model_eva.y.shape == (12, 1)

        # Clear test report dir
        if os.path.exists(temp_dir):
            shutil.rmtree(temp_dir)

    # Classification model evaluation tests
    @staticmethod
    def test_classifier_data_split() -> None:
        """Test data split functionality."""
        with tempfile.TemporaryDirectory() as temp_dir:
            # K-fold
            model_eva = modeleva_initialization_classification(temp_dir, data_split="5-fold", n_samples=20)

            assert model_eva.dsp_inds == []
            model_eva._data_split()
            assert len(model_eva.dsp_inds) == 5
            assert len(model_eva.dsp_inds[0][0]) == 16
            assert len(model_eva.dsp_inds[0][1]) == 4

            # LOO
            model_eva = modeleva_initialization_classification(temp_dir, data_split="loo")

            assert model_eva.dsp_inds == []
            model_eva._data_split()
            assert len(model_eva.dsp_inds) == 10
            assert len(model_eva.dsp_inds[0][0]) == 9
            assert len(model_eva.dsp_inds[0][1]) == 1

            # Train-test-split
            model_eva = modeleva_initialization_classification(temp_dir, data_split="60-40-split")

            assert model_eva.dsp_inds == []
            model_eva._data_split()
            assert len(model_eva.dsp_inds) == 1
            assert len(model_eva.dsp_inds[0][0]) == 6
            assert len(model_eva.dsp_inds[0][1]) == 4

        # Clear test report dir
        if os.path.exists(temp_dir):
            shutil.rmtree(temp_dir)

    @staticmethod
    def test_classifier_validation() -> None:
        """Test data split functionality."""
        with tempfile.TemporaryDirectory() as temp_dir:
            model_eva = modeleva_initialization_classification(temp_dir, n_samples=20)
            model_eva._data_split()
            model_time = model_eva.model_time

            assert len(model_time) > 0
            assert model_eva.y_true_eva is None
            assert model_eva.y_pred_eva is None
            assert model_eva.y_true_proba_eva is None
            assert model_eva.y_pred_proba_eva is None
            assert model_eva.sid_eva == []

            model_eva._classifier_validation()

            assert len(model_time) > 0
            assert model_eva.y_true_eva is not None
            assert model_eva.y_pred_eva is not None
            assert model_eva.y is not None
            assert len(model_eva.y_true_eva) == len(model_eva.y)
            assert len(model_eva.y_pred_eva) == len(model_eva.y)
            assert model_eva.y_true_proba_eva is not None
            assert np.all(model_eva.y_true_proba_eva.shape == (len(model_eva.y), len(model_eva.ynames)))
            assert model_eva.y_pred_proba_eva is not None
            assert np.all(model_eva.y_pred_proba_eva.shape == (len(model_eva.y), len(model_eva.ynames)))
            assert len(model_eva.sid_eva) > 0

        # Clear test report dir
        if os.path.exists(temp_dir):
            shutil.rmtree(temp_dir)

    @staticmethod
    def test_classifier_evaluation() -> None:
        """Test classifier evaluation workflow."""
        with tempfile.TemporaryDirectory() as temp_dir:
            model_eva = modeleva_initialization_classification(temp_dir, data_split="10-fold", n_samples=20)
            model_eva.classifier_evaluation()
            metrics_filename = f"Classification_performance_{model_eva.model_label}.csv"
            roc_filename = f"ROC_curve_{model_eva.model_label}.png"
            residual_filename = f"Residual_analysis_{model_eva.model_label}.csv"
            influence_analysis_filename = f"Influence_analysis_{model_eva.model_label}.csv"
            output_path = (
                model_eva._report_directory
                + f"Model_evaluation_reports/Data_{model_eva.data_label}_Model_{model_eva.model_label}/"
            )

            # Test metrics report output
            metrics_path = output_path + metrics_filename
            assert os.path.exists(metrics_path)
            try:
                metrics = pd.read_csv(metrics_path)
                assert True
            except Exception as e:
                raise AssertionError(f"Failed to read CSV file '{metrics_filename}': {e}") from e
            assert list(metrics.columns) == [
                "Class",
                "TP",
                "TN",
                "FP",
                "FN",
                "Precision",
                "Recall",
                "F1_Score",
                "Accuracy",
                "AUC",
            ]
            assert "LogisticRegression" in metrics.iloc[-1, 0]
            assert metrics.shape == (5, 10)
            assert metrics.iloc[:-1, :].isna().sum().sum() == 0

            # Test ROC plot existence (manual check for plot image required)
            roc_path = output_path + roc_filename
            assert os.path.exists(roc_path)

            # Test residual report output
            residual_path = output_path + residual_filename
            assert os.path.exists(residual_path)
            try:
                residual = pd.read_csv(residual_path)
                assert True
            except Exception as e:
                raise AssertionError(f"Failed to read CSV file '{residual_filename}': {e}") from e
            assert residual.shape == (20, 15)
            assert residual.iloc[:, :].isna().sum().sum() == 0

            # Test Cook's distance report output
            influence_analysis_path = output_path + influence_analysis_filename
            assert os.path.exists(influence_analysis_path)
            try:
                influence = pd.read_csv(influence_analysis_path)
                assert True
            except Exception as e:
                raise AssertionError(f"Failed to read CSV file '{influence_analysis_filename}': {e}") from e
            assert list(influence.columns) == [
                "Sample_ID",
                "A_probability_Cooks_distance_like",
                "B_probability_Cooks_distance_like",
            ]
            assert influence.shape == (20, 3)
            assert influence.iloc[:, :].isna().sum().sum() == 0

        # Clear test report dir
        if os.path.exists(temp_dir):
            shutil.rmtree(temp_dir)

    @staticmethod
    def test_classifier_application_model() -> None:
        """Test application model training and application and validation model file existence."""
        with tempfile.TemporaryDirectory() as temp_dir:
            # Test train_application_model
            model_eva = modeleva_initialization_classification(temp_dir, data_split="60-40-split")

            model_eva._data_split()
            model_eva._classifier_validation()
            assert model_eva.app_model is None

            model_eva.train_application_model()
            assert type(model_eva.app_model) is LogisticRegression

            # Test default app_model training in classifier_evaluation
            model_eva = modeleva_initialization_classification(temp_dir, data_split="60-40-split")
            assert model_eva.app_model is None
            model_eva.classifier_evaluation()
            assert type(model_eva.app_model) is LogisticRegression

            # Test saved models
            val_model_path = (
                model_eva._report_directory
                + "Model_evaluation_reports/"
                + f"Data_{model_eva.data_label}_Model_{model_eva.model_label}/Model_in_validation/"
                + f"val_model_fold-0_{model_eva._model_label}.dill"
            )
            assert os.path.exists(val_model_path)
            app_model_path = (
                model_eva._report_directory
                + "Model_evaluation_reports/"
                + f"Data_{model_eva.data_label}_Model_{model_eva.model_label}/Model_for_application/"
                + f"app_model_{model_eva._model_label}.dill"
            )
            assert os.path.exists(app_model_path)

        # Clear test report dir
        if os.path.exists(temp_dir):
            shutil.rmtree(temp_dir)

    # Regression model evaluation tests
    @staticmethod
    def test_regressor_data_split() -> None:
        """Test data split functionality."""
        with tempfile.TemporaryDirectory() as temp_dir:
            # K-fold
            model_eva = modeleva_initialization_regression(temp_dir, data_split="5-fold")

            assert model_eva.dsp_inds == []
            model_eva._data_split()
            assert len(model_eva.dsp_inds) == 5
            assert len(model_eva.dsp_inds[0][0]) == 8
            assert len(model_eva.dsp_inds[0][1]) == 2

            # LOO
            model_eva = modeleva_initialization_regression(temp_dir, data_split="loo")

            assert model_eva.dsp_inds == []
            model_eva._data_split()
            assert len(model_eva.dsp_inds) == 10
            assert len(model_eva.dsp_inds[0][0]) == 9
            assert len(model_eva.dsp_inds[0][1]) == 1

            # Train-test-split
            model_eva = modeleva_initialization_regression(temp_dir, data_split="60-40-split")

            assert model_eva.dsp_inds == []
            model_eva._data_split()
            assert len(model_eva.dsp_inds) == 1
            assert len(model_eva.dsp_inds[0][0]) == 6
            assert len(model_eva.dsp_inds[0][1]) == 4

        # Clear test report dir
        if os.path.exists(temp_dir):
            shutil.rmtree(temp_dir)

    @staticmethod
    def test_regressor_validation() -> None:
        """Test data split functionality."""
        with tempfile.TemporaryDirectory() as temp_dir:
            model_eva = modeleva_initialization_regression(temp_dir)
            model_eva._data_split()
            model_time = model_eva.model_time

            assert len(model_time) > 0
            assert model_eva.y_true_eva is None
            assert model_eva.y_pred_eva is None
            assert model_eva.y_true_proba_eva is None
            assert model_eva.y_pred_proba_eva is None
            assert model_eva.sid_eva == []

            model_eva._regressor_validation()

            assert len(model_time) > 0
            assert model_eva.y is not None
            assert model_eva.sid_eva is not None
            assert model_eva.y_true_eva is not None
            assert model_eva.y_pred_eva is not None
            assert model_eva.y_true_proba_eva is None
            assert model_eva.y_pred_proba_eva is None
            assert len(model_eva.y_true_eva) == len(model_eva.y)
            assert len(model_eva.y_pred_eva) == len(model_eva.y)
            assert len(model_eva.sid_eva) > 0

        # Clear test report dir
        if os.path.exists(temp_dir):
            shutil.rmtree(temp_dir)

    @staticmethod
    def test_regressor_evaluation() -> None:
        """Test classifier evaluation workflow."""
        with tempfile.TemporaryDirectory() as temp_dir:
            model_eva = modeleva_initialization_regression(temp_dir, data_split="5-fold")
            model_eva.regressor_evaluation()
            metrics_filename = f"Regression_performance_{model_eva.model_label}.csv"
            scatter_plot_filename = f"Scatter_plot_{model_eva.model_label}.png"
            residual_plot_filename = f"Residual_plot_{model_eva.model_label}.png"
            residual_filename = f"Residual_analysis_{model_eva.model_label}.csv"
            influence_analysis_filename = f"Influence_analysis_{model_eva.model_label}.csv"
            output_path = (
                model_eva._report_directory
                + f"Model_evaluation_reports/Data_{model_eva.data_label}_Model_{model_eva.model_label}/"
            )

            # Test metrics report output
            metrics_path = output_path + metrics_filename
            assert os.path.exists(metrics_path)
            try:
                metrics = pd.read_csv(metrics_path)
                assert True
            except Exception as e:
                raise AssertionError(f"Failed to read CSV file '{metrics_filename}': {e}") from e
            assert list(metrics.columns) == [
                "Mean_Error",
                "Standard_Deviation_of_Error",
                "Mean_Absolute_Error",
                "Normalized_MAE",
                "CV_MAE",
                "Mean_Squared_Error",
                "Root_Mean_Squared_Error",
                "Normalized_RMSE",
                "CV_RMSE",
                "Residual_Prediction_Deviation",
                "R2",
            ]
            assert metrics.shape == (1, 11)
            assert metrics.iloc[:, :].isna().sum().sum() == 0

            # Test scatter plot and residual plot existence (manual check for plot image required)
            scatter_plot_path = output_path + scatter_plot_filename
            assert os.path.exists(scatter_plot_path)
            residual_plot_path = output_path + residual_plot_filename
            assert os.path.exists(residual_plot_path)

            # Test residual report output
            residual_path = output_path + residual_filename
            assert os.path.exists(residual_path)
            try:
                residual = pd.read_csv(residual_path)
                assert True
            except Exception as e:
                raise AssertionError(f"Failed to read CSV file '{residual_filename}': {e}") from e
            assert residual.shape == (10, 5)
            assert residual.iloc[:, :].isna().sum().sum() == 0

            # Test Cook's distance report output
            influence_analysis_path = output_path + influence_analysis_filename
            assert os.path.exists(influence_analysis_path)
            try:
                influence = pd.read_csv(influence_analysis_path)
                assert True
            except Exception as e:
                raise AssertionError(f"Failed to read CSV file '{influence_analysis_filename}': {e}") from e
            assert list(influence.columns) == ["Sample_ID", "Cooks_distance_like"]
            assert influence.shape == (10, 2)
            assert influence.iloc[:, :].isna().sum().sum() == 0

        # Clear test report dir
        if os.path.exists(temp_dir):
            shutil.rmtree(temp_dir)

    @staticmethod
    def test_regressor_application_model() -> None:
        """Test application model training and application and validation model file existence."""
        with tempfile.TemporaryDirectory() as temp_dir:
            # Test train_application_model
            model_eva = modeleva_initialization_regression(temp_dir, data_split="60-40-split")

            model_eva._data_split()
            model_eva._regressor_validation()
            assert model_eva.app_model is None

            model_eva.train_application_model()
            assert type(model_eva.app_model) is LinearRegression

            # Test default app_model training in regressor_evaluation
            model_eva = modeleva_initialization_regression(temp_dir, data_split="60-40-split")
            assert model_eva.app_model is None
            model_eva.regressor_evaluation()
            assert type(model_eva.app_model) is LinearRegression

            # Test saved models
            val_model_path = (
                model_eva._report_directory
                + "Model_evaluation_reports/"
                + f"Data_{model_eva.data_label}_Model_{model_eva.model_label}/Model_in_validation/"
                + f"val_model_fold-0_{model_eva._model_label}.dill"
            )
            assert os.path.exists(val_model_path)
            app_model_path = (
                model_eva._report_directory
                + "Model_evaluation_reports/"
                + f"Data_{model_eva.data_label}_Model_{model_eva.model_label}/Model_for_application/"
                + f"app_model_{model_eva._model_label}.dill"
            )
            assert os.path.exists(app_model_path)

        # Clear test report dir
        if os.path.exists(temp_dir):
            shutil.rmtree(temp_dir)


# %% Tests - ModelEva


# TestModelEva.test_initialization_regression()
# TestModelEva.test_initialization_classification()
# TestModelEva.test_auto_choose_classification_regression()
# TestModelEva.test_auto_kfold_to_loocv()

# TestModelEva.test_property_access()

# TestModelEva.test_update_samples()

# TestModelEva.test_classifier_data_split()
# TestModelEva.test_classifier_validation()
# TestModelEva.test_classifier_evaluation()
# TestModelEva.test_classifier_application_model()

# TestModelEva.test_regressor_data_split()
# TestModelEva.test_regressor_validation()
# TestModelEva.test_regressor_evaluation()
# TestModelEva.test_regressor_application_model()


# %% Test main

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
Tests for raster operation functions (RasterOp)

The following source code was created with AI assistance and has been human reviewed and edited.

Copyright (c) 2025 Siwei Luo. MIT License.
"""

# Test
import pytest
import unittest

# OS Files
import os
import tempfile

# Typing
from typing import Union

# Testing third
import numpy as np
import torch

# Rasters
import rasterio

# Self
# Applied package functions for test
from specpipe.example_data import create_test_raster

# Funcs to test
from specpipe.rasterop import (
    auto_fp,
    croproi,
    dtype_mapper,
    pixel_apply,
    pixel_array_apply,
    pixel_spec_apply,
    pixel_tensor_apply,
    pixel_tensor_hyper_apply,
)

# Check if cuda is available
try:
    HAS_CUDA = torch.cuda.is_available()
except ImportError:
    HAS_CUDA = False

# %% test functions : croproi


class TestCropROI(unittest.TestCase):
    @staticmethod
    def test_croproi_basic_functionality() -> None:
        """Test basic cropping functionality with a simple rectangle ROI."""
        with tempfile.TemporaryDirectory() as temp_dir:
            # Create test raster with 50x100 dimensions and 100 bands
            raster_path = os.path.join(temp_dir, "test_raster.tif")
            output_path = os.path.join(temp_dir, "cropped.tif")

            create_test_raster(raster_path, width=100, height=50, bands=4)

            # Define ROI coordinates (a rectangle in the center)
            roi_coordinates = [
                [
                    (25.0, 10.0),  # Top-left
                    (75.0, 10.0),  # Top-right
                    (75.0, 40.0),  # Bottom-right
                    (25.0, 40.0),  # Bottom-left
                    (25.0, 10.0),  # Close the polygon
                ]
            ]

            # Call the function
            croproi(raster_path, roi_coordinates, output_path)

            # Verify output file exists
            assert os.path.exists(output_path)

            # Verify output properties
            with rasterio.open(output_path) as src:
                # Should be cropped to 50x30 (from 25-75 in width, 10-40 in height)
                assert src.width == 50
                assert src.height == 30
                assert src.count == 4  # All 100 bands preserved

                # Read some sample data to verify it's not all zeros
                data = src.read(1)  # Read first band
                assert not np.all(data == 0)  # Should have non-zero values

    @staticmethod
    def test_croproi_multiple_polygons() -> None:
        """Test cropping with multiple ROI polygons."""
        with tempfile.TemporaryDirectory() as temp_dir:
            raster_path = os.path.join(temp_dir, "test_raster.tif")
            output_path = os.path.join(temp_dir, "cropped_multi.tif")

            # Create test raster with 50x100 dimensions and 120 bands
            create_test_raster(raster_path, width=100, height=50, bands=4)

            # Define two separate ROI polygons
            roi_coordinates = [
                [  # First polygon (top-left corner)
                    (0.0, 0.0),
                    (25.0, 0.0),
                    (25.0, 15.0),
                    (0.0, 15.0),
                    (0.0, 0.0),
                ],
                [  # Second polygon (bottom-right corner)
                    (75.0, 35.0),
                    (100.0, 35.0),
                    (100.0, 50.0),
                    (75.0, 50.0),
                    (75.0, 35.0),
                ],
            ]

            croproi(raster_path, roi_coordinates, output_path)

            assert os.path.exists(output_path)

            with rasterio.open(output_path) as src:
                # Should contain both regions with all bands
                assert src.width > 0
                assert src.height > 0
                assert src.count == 4  # All bands preserved

    @staticmethod
    def test_croproi_high_band_count() -> None:
        """Test cropping with a very high number of bands."""
        with tempfile.TemporaryDirectory() as temp_dir:
            raster_path = os.path.join(temp_dir, "test_raster.tif")
            output_path = os.path.join(temp_dir, "cropped_high_bands.tif")

            # Create test raster with 50x100 dimensions and 224 bands (like AVIRIS)
            create_test_raster(raster_path, width=100, height=50, bands=224)

            roi_coordinates = [[(20.0, 10.0), (80.0, 10.0), (80.0, 40.0), (20.0, 40.0), (20.0, 10.0)]]

            croproi(raster_path, roi_coordinates, output_path)

            assert os.path.exists(output_path)

            with rasterio.open(output_path) as src:
                assert src.count == 224  # All 224 bands preserved
                assert src.width == 60  # 80-20 = 60
                assert src.height == 30  # 40-10 = 30

    @staticmethod
    def test_croproi_complex_polygons() -> None:
        """Test cropping with complex polygon shapes."""
        with tempfile.TemporaryDirectory() as temp_dir:
            raster_path = os.path.join(temp_dir, "test_raster.tif")
            output_path = os.path.join(temp_dir, "cropped_complex.tif")

            # Create test raster with 50x100 dimensions and 150 bands
            create_test_raster(raster_path, width=100, height=50, bands=150)

            # Complex polygon with multiple vertices
            roi_coordinates = [
                [
                    (10.0, 5.0),
                    (40.0, 5.0),
                    (60.0, 15.0),
                    (80.0, 5.0),
                    (90.0, 20.0),
                    (80.0, 35.0),
                    (90.0, 45.0),
                    (70.0, 45.0),
                    (50.0, 35.0),
                    (30.0, 45.0),
                    (15.0, 35.0),
                    (5.0, 20.0),
                    (10.0, 5.0),
                ]
            ]

            croproi(raster_path, roi_coordinates, output_path)

            assert os.path.exists(output_path)

            with rasterio.open(output_path) as src:
                assert src.count == 150  # All bands preserved
                assert src.width > 0
                assert src.height > 0

    @staticmethod
    def test_croproi_metadata_and_band_integrity() -> None:
        """Test that all bands and metadata are properly preserved."""
        with tempfile.TemporaryDirectory() as temp_dir:
            raster_path = os.path.join(temp_dir, "test_raster.tif")
            output_path = os.path.join(temp_dir, "cropped_meta.tif")

            # Create test raster with 50x100 dimensions and 200 bands
            create_test_raster(raster_path, width=100, height=50, bands=200)

            roi_coordinates = [[(20.0, 10.0), (80.0, 10.0), (80.0, 40.0), (20.0, 40.0), (20.0, 10.0)]]

            croproi(raster_path, roi_coordinates, output_path)

            with rasterio.open(raster_path) as src_orig:
                with rasterio.open(output_path) as src_cropped:
                    # Check that all metadata and bands are preserved
                    assert src_orig.count == src_cropped.count
                    assert src_orig.dtypes[0] == src_cropped.dtypes[0]
                    assert src_orig.crs == src_cropped.crs
                    assert src_orig.nodata == src_cropped.nodata

                    # Verify band names or descriptions if they exist
                    if hasattr(src_orig, "descriptions") and src_orig.descriptions[0]:
                        assert src_orig.descriptions == src_cropped.descriptions

    @staticmethod
    def test_croproi_edge_cases() -> None:
        """Test edge cases with different band counts and ROI positions."""
        with tempfile.TemporaryDirectory() as temp_dir:
            # Test with exactly 100 bands
            raster_path = os.path.join(temp_dir, "test_raster_100.tif")
            output_path = os.path.join(temp_dir, "cropped_100.tif")

            create_test_raster(raster_path, width=100, height=50, bands=100)

            # ROI covering most of the image
            roi_coordinates = [[(5.0, 5.0), (95.0, 5.0), (95.0, 45.0), (5.0, 45.0), (5.0, 5.0)]]

            croproi(raster_path, roi_coordinates, output_path)

            with rasterio.open(output_path) as src:
                assert src.count == 100
                assert src.width == 90  # 95-5 = 90
                assert src.height == 40  # 45-5 = 40


# %% Test - croproi

# TestCropROI.test_croproi_basic_functionality()
# TestCropROI.test_croproi_multiple_polygons()
# TestCropROI.test_croproi_high_band_count()
# TestCropROI.test_croproi_complex_polygons()
# TestCropROI.test_croproi_metadata_and_band_integrity()
# TestCropROI.test_croproi_edge_cases()


# %% test functions : auto_fp


class TestAutoFP(unittest.TestCase):
    @staticmethod
    def test_numpy_short_types_preserved() -> None:
        """Test that numpy short types are preserved without conversion."""
        # Test int8
        data_int8 = np.array([1, 2, 3], dtype=np.int8)
        result, scale = auto_fp(data_int8)
        assert result.dtype == np.int8
        assert scale is None

        # Test uint8
        data_uint8 = np.array([1, 2, 3], dtype=np.uint8)
        result, scale = auto_fp(data_uint8)
        assert result.dtype == np.uint8
        assert scale is None

        # Test float16
        data_fp16 = np.array([1.0, 2.0, 3.0], dtype=np.float16)
        result, scale = auto_fp(data_fp16)
        assert result.dtype == np.float16
        assert scale is None

    @staticmethod
    def test_numpy_long_types_compatible_fp16() -> None:
        """Test numpy long types that fit natively in FP16."""
        # Float32 data within FP16 range
        data = np.array([1.0, 2.0, 3.0], dtype=np.float32)
        result, scale = auto_fp(data)
        assert result.dtype == np.float16
        assert scale == 1.0

        # Float64 data within FP16 range
        data = np.array([1.0, 2.0, 3.0], dtype=np.float64)
        result, scale = auto_fp(data)
        assert result.dtype == np.float16
        assert scale == 1.0

    @staticmethod
    def test_numpy_scaling_large_values() -> None:
        """Test scaling of large numpy values to fit FP16."""
        # Create data that exceeds FP16 max
        large_data = np.array([100000, 200000], dtype=np.float32)

        result, scale = auto_fp(large_data, scaling=True)
        assert result is not None
        assert scale is not None
        assert result.dtype == np.float16
        assert scale < 1.0  # Should be a scaling down factor

        # Verify scaling was applied correctly
        expected_scaled = large_data * scale
        assert type(result) is np.ndarray
        assert np.allclose(result.astype(np.float32), expected_scaled, rtol=1e-4, atol=1e-5)

    @staticmethod
    def test_numpy_scaling_small_values() -> None:
        """Test scaling of small numpy values to fit FP16."""
        # Create data with values below FP16 min
        small_data = np.array([2e-7, 1e-7], dtype=np.float64)

        result, scale = auto_fp(small_data, scaling=True)
        assert result is not None
        assert scale is not None
        assert result.dtype == np.float16
        assert scale > 1.0  # Should be a scaling up factor

        # Verify scaling was applied correctly
        expected_scaled = small_data * scale
        assert np.allclose(result.astype(np.float64), expected_scaled, rtol=1e-4, atol=1e-5)

    @staticmethod
    def test_numpy_precision_loss_rejection() -> None:
        """Test that numpy data with precision loss is rejected."""
        # Create data that would lose precision when scaled
        problematic_data = np.array([1.23456789, 2.34567891], dtype=np.float64)

        result, scale = auto_fp(problematic_data, scaling=True, rtol=1e-8, atol=1e-8)
        assert result.dtype == np.float64  # Should return original
        assert scale is None

    @staticmethod
    def test_torch_fp16_preserved() -> None:
        """Test that torch float16 tensors are preserved."""
        data = torch.tensor([1.0, 2.0, 3.0], dtype=torch.float16)
        result, scale = auto_fp(data)
        assert result.dtype == torch.float16
        assert scale is None

    @staticmethod
    def test_torch_float32_compatible_fp16() -> None:
        """Test torch float32 tensors that fit natively in FP16."""
        data = torch.tensor([1.0, 2.0, 3.0], dtype=torch.float32)
        result, scale = auto_fp(data)
        assert result is not None
        assert scale is not None
        assert result.dtype == torch.float16
        assert scale == 1.0

    @staticmethod
    def test_torch_scaling_large_values() -> None:
        """Test scaling of large torch tensor values to fit FP16."""
        # Create data that exceeds FP16 max
        large_data = torch.tensor([100000, 200000], dtype=torch.float32)

        result, scale = auto_fp(large_data, scaling=True)
        assert result is not None
        assert scale is not None
        assert result.dtype == torch.float16
        assert scale < 1.0

        # Verify scaling was applied correctly
        expected_scaled = large_data * scale
        assert torch.allclose(result.to(torch.float32), expected_scaled, rtol=1e-4, atol=1e-5)

    @staticmethod
    def test_torch_scaling_small_values() -> None:
        """Test scaling of small torch tensor values to fit FP16."""
        # Create data with values below FP16 min
        small_data = torch.tensor([2e-7, 1e-7], dtype=torch.float64)

        result, scale = auto_fp(small_data, scaling=True)
        assert result is not None
        assert scale is not None
        assert result.dtype == torch.float16
        assert scale > 1.0

        # Verify scaling was applied correctly
        expected_scaled = small_data * scale
        assert torch.allclose(result.to(torch.float64), expected_scaled, rtol=1e-4, atol=1e-5)

    @staticmethod
    def test_torch_no_scaling_option() -> None:
        """Test that torch tensors return original when scaling=False and incompatible."""
        fp16_max = 65504 / 2.0
        large_data = torch.tensor([fp16_max * 2], dtype=torch.float32)

        result, scale = auto_fp(large_data, scaling=False)
        assert result.dtype == torch.float32  # Should return original
        assert scale is None

    @staticmethod
    def test_torch_precision_loss_rejection() -> None:
        """Test that torch data with precision loss is rejected."""
        problematic_data = torch.tensor([1.23456789, 2.34567891], dtype=torch.float64)

        result, scale = auto_fp(problematic_data, scaling=True, rtol=1e-8, atol=1e-8)
        assert result.dtype == torch.float64  # Should return original
        assert scale is None

    @staticmethod
    def test_unsupported_data_type() -> None:
        """Test that unsupported data types raise ValueError."""
        # Test with complex numbers (unsupported)
        complex_data = np.array([1 + 2j, 3 + 4j], dtype=np.complex64)

        try:
            auto_fp(complex_data)
            raise AssertionError("Should have raised ValueError")
        except ValueError as e:
            assert "Unsupported data dtype" in str(e)

    @staticmethod
    def test_edge_case_zero_values() -> None:
        """Test handling of zero values which should be excluded from min_val calculation."""
        data_with_zeros = np.array([0.0, 1e-6, 0.001], dtype=np.float32)
        result, scale = auto_fp(data_with_zeros, scaling=True)

        # Should handle zeros gracefully without affecting scaling
        assert result.dtype == np.float16
        assert scale is not None

    @staticmethod
    def test_custom_tolerance_parameters() -> None:
        """Test that custom rtol and atol parameters are respected."""
        # Create data that would pass with default tolerance but fail with stricter
        borderline_data = np.array([1.0001, 2.0001], dtype=np.float64)

        # Should pass with default tolerance
        result1, scale1 = auto_fp(borderline_data, scaling=True, rtol=1e-3, atol=1e-3)
        assert result1.dtype == np.float16

        # Should fail with stricter tolerance
        result2, scale2 = auto_fp(borderline_data, scaling=True, rtol=1e-8, atol=1e-8)
        assert result2.dtype == np.float64
        assert scale2 is None

    @staticmethod
    def test_custom_safety_factor() -> None:
        """Test that custom safety_factor affects FP16 limits."""
        custom_safety = 1.1
        fp16_max_custom = 40000.0

        # Data that fits with custom safety but not with default
        data = np.array([fp16_max_custom], dtype=np.float32)

        # Should convert with custom safety factor
        result, scale = auto_fp(data, safety_factor=custom_safety)
        assert result.dtype == np.float16
        assert scale == 1.0

        # Should not convert with default safety factor
        result_default, scale_default = auto_fp(data, safety_factor=2.0)
        assert result_default.dtype == np.float32
        assert scale_default is None


# %% Test - auto_fp

# TestAutoFP.test_numpy_short_types_preserved()
# TestAutoFP.test_numpy_long_types_compatible_fp16()

# TestAutoFP.test_numpy_scaling_large_values()
# TestAutoFP.test_numpy_scaling_small_values()

# TestAutoFP.test_numpy_precision_loss_rejection()

# TestAutoFP.test_torch_fp16_preserved()
# TestAutoFP.test_torch_float32_compatible_fp16()

# TestAutoFP.test_torch_scaling_large_values()
# TestAutoFP.test_torch_scaling_small_values()

# TestAutoFP.test_torch_no_scaling_option()

# TestAutoFP.test_torch_precision_loss_rejection()

# TestAutoFP.test_unsupported_data_type()
# TestAutoFP.test_edge_case_zero_values()

# TestAutoFP.test_custom_tolerance_parameters()
# TestAutoFP.test_custom_safety_factor()


# %% test functions : dtype_mapper


def test_dtype_mapper() -> None:
    """Test dtype_mapper functionality"""

    # Test basic numpy mappings
    assert dtype_mapper("float32", "numpy") == np.float32
    assert dtype_mapper("int16", "numpy") == np.int16
    assert dtype_mapper("uint8", "numpy") == np.uint8
    assert dtype_mapper("bool", "numpy") == np.bool_

    # Test numpy type inputs
    assert dtype_mapper(np.float32, "numpy") == np.float32
    assert dtype_mapper(np.int64, "numpy") == np.int64

    # Test Python native types
    assert dtype_mapper(float, "numpy") == np.float32
    assert dtype_mapper(int, "numpy") == np.int32
    assert dtype_mapper(bool, "numpy") == np.bool_

    # Test raster mappings
    assert dtype_mapper("float32", "raster") == "float32"
    assert dtype_mapper("int16", "raster") == "int16"
    assert dtype_mapper("uint8", "raster") == "uint8"
    assert dtype_mapper(float, "raster") == "float32"
    assert dtype_mapper(int, "raster") == "int32"

    # Test raster compatibility mappings (min_compatible=True)
    assert dtype_mapper("float16", "raster") == "float32"
    assert dtype_mapper("int8", "raster") == "int16"

    # Test torch mappings
    assert dtype_mapper("float32", "torch") == torch.float32
    assert dtype_mapper("int32", "torch") == torch.int32
    assert dtype_mapper("uint8", "torch") == torch.uint8

    # Test torch compatibility mappings
    assert dtype_mapper("uint16", "torch") == torch.int32
    assert dtype_mapper(np.uint16, "torch") == torch.int32

    # Test various string representations
    assert dtype_mapper("double", "numpy") == np.float64
    assert dtype_mapper("half", "numpy") == np.float16
    assert dtype_mapper("int", "numpy") == np.int32
    assert dtype_mapper("float", "numpy") == np.float32

    # Test raster mappings for string representations
    assert dtype_mapper("double", "raster") == "float32"
    assert dtype_mapper("half", "raster") == "float32"

    # Test error cases
    with pytest.raises(ValueError, match="Unsupported map_type"):
        dtype_mapper("float32", "invalid_type")

    with pytest.raises(ValueError, match="Unsupported NumPy data type"):
        dtype_mapper("invalid_dtype", "numpy")

    # Test that identical types are returned unchanged
    assert dtype_mapper(np.float32, "numpy") is np.float32
    assert dtype_mapper("float32", "raster") == "float32"


# %% Test - dtype_mapper

# test_dtype_mapper()


# %% test functions : pixel_spec_apply


class TestPixelSpecApply:
    """Test class for pixel_spec_apply function"""

    @staticmethod
    def simple_spectral_function(spectrum: np.ndarray) -> np.ndarray:
        """Simple test spectral function that doubles the values"""
        return spectrum * 2

    @staticmethod
    def spectral_function_with_different_output(spectrum: np.ndarray) -> np.ndarray:
        """Spectral function that returns different number of bands"""
        result: np.ndarray = np.array([np.sum(spectrum), np.mean(spectrum), np.max(spectrum)])
        return result

    @staticmethod
    def test_basic_functionality() -> None:
        """Test basic functionality with a simple spectral function"""
        with tempfile.TemporaryDirectory() as temp_dir:
            # Create test files
            input_path = os.path.join(temp_dir, "input.tif")
            output_path = os.path.join(temp_dir, "output.tif")

            # Create test raster
            create_test_raster(input_path, 8, 8, 10)

            # Apply function
            pixel_spec_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelSpecApply.simple_spectral_function,
                tile_size=4,
                progress=False,
            )

            # Verify output exists
            assert os.path.exists(output_path)

            # Verify output content
            with rasterio.open(input_path) as src_input, rasterio.open(output_path) as src_output:
                # Check metadata
                assert src_output.count == src_input.count  # Same number of bands
                assert src_output.width == src_input.width
                assert src_output.height == src_input.height

                # Check data content
                input_data = src_input.read()
                output_data = src_output.read()

                # Verify the spectral function was applied correctly
                expected_data = input_data * 2
                np.testing.assert_array_almost_equal(output_data, expected_data)

    @staticmethod
    def test_different_output_bands() -> None:
        """Test with spectral function that returns different number of bands"""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "input.tif")
            output_path = os.path.join(temp_dir, "output.tif")

            create_test_raster(input_path, 9, 9, 4)

            pixel_spec_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelSpecApply.spectral_function_with_different_output,
                tile_size=3,
                progress=False,
            )

            assert os.path.exists(output_path)

            with rasterio.open(output_path) as src:
                assert src.count == 3  # Should have 3 output bands

    @staticmethod
    def test_different_dtypes() -> None:
        """Test with different output dtypes"""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "input.tif")
            output_path = os.path.join(temp_dir, "output.tif")

            create_test_raster(input_path, 4, 4, 2, dtype="uint8")

            pixel_spec_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelSpecApply.simple_spectral_function,
                dtype="float64",
                tile_size=2,
                progress=False,
            )

            assert os.path.exists(output_path)

            with rasterio.open(output_path) as src:
                assert src.dtypes[0] == "float64"

    @staticmethod
    def test_default_tile_size() -> None:
        """Test with default tile size (-1)"""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "input.tif")
            output_path = os.path.join(temp_dir, "output.tif")

            create_test_raster(input_path, 10, 10, 4)

            pixel_spec_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelSpecApply.simple_spectral_function,
                tile_size=-1,  # Should use default
                progress=False,
            )

            assert os.path.exists(output_path)

    @staticmethod
    def test_invalid_tile_size() -> None:
        """Test with invalid tile size"""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "input.tif")
            output_path = os.path.join(temp_dir, "output.tif")

            create_test_raster(input_path, 4, 4, 2)

            with pytest.raises(ValueError, match="If provided, tile_size must be positive"):
                pixel_spec_apply(
                    image_path=input_path,
                    output_path=output_path,
                    spectral_function=TestPixelSpecApply.simple_spectral_function,
                    tile_size=0,  # Invalid
                    progress=False,
                )

    @staticmethod
    def test_output_file_overwrite() -> None:
        """Test that existing output file is overwritten"""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "input.tif")
            output_path = os.path.join(temp_dir, "output.tif")

            create_test_raster(input_path, 4, 4, 2)

            # Create output file first
            with open(output_path, "w") as f:
                f.write("test content")

            assert os.path.exists(output_path)

            # Should overwrite without error
            pixel_spec_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelSpecApply.simple_spectral_function,
                tile_size=2,
                progress=False,
            )

            # Should be a valid raster file now
            assert os.path.exists(output_path)
            with rasterio.open(output_path) as src:
                assert src.count == 2

    @staticmethod
    def test_edge_case_tile_sizes() -> None:
        """Test with tile sizes that don't evenly divide image dimensions"""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "input.tif")
            output_path = os.path.join(temp_dir, "output.tif")

            create_test_raster(input_path, 7, 5, 3)  # Odd dimensions

            pixel_spec_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelSpecApply.simple_spectral_function,
                tile_size=3,  # Doesn't divide evenly
                progress=False,
            )

            assert os.path.exists(output_path)

            # Verify all pixels were processed
            with rasterio.open(input_path) as src_input, rasterio.open(output_path) as src_output:
                input_data = src_input.read()
                output_data = src_output.read()

                expected_data = input_data * 2
                np.testing.assert_array_almost_equal(output_data, expected_data)

    @staticmethod
    def test_spectral_function_validation() -> None:
        """Test that spectral function is properly validated"""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "input.tif")
            output_path = os.path.join(temp_dir, "output.tif")

            create_test_raster(input_path, 2, 2, 2)

            # Test with invalid function (wrong signature)
            def invalid_function(x: Union[int, float], y: Union[int, float]) -> Union[int, float]:
                return x + y

            with pytest.raises((TypeError, ValueError)):
                pixel_spec_apply(
                    image_path=input_path,
                    output_path=output_path,
                    spectral_function=invalid_function,
                    tile_size=2,
                    progress=False,
                )

    @staticmethod
    def test_single_pixel_image() -> None:
        """Test with single pixel image"""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "input.tif")
            output_path = os.path.join(temp_dir, "output.tif")

            create_test_raster(input_path, 1, 1, 4)

            pixel_spec_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelSpecApply.simple_spectral_function,
                tile_size=1,
                progress=False,
            )

            assert os.path.exists(output_path)

            with rasterio.open(input_path) as src_input, rasterio.open(output_path) as src_output:
                input_data = src_input.read()
                output_data = src_output.read()

                expected_data = input_data * 2
                np.testing.assert_array_almost_equal(output_data, expected_data)


# %% Test - pixel_spec_apply

# TestPixelSpecApply.test_basic_functionality()
# TestPixelSpecApply.test_different_output_bands()
# TestPixelSpecApply.test_different_dtypes()
# TestPixelSpecApply.test_default_tile_size()
# TestPixelSpecApply.test_invalid_tile_size()
# TestPixelSpecApply.test_output_file_overwrite()
# TestPixelSpecApply.test_edge_case_tile_sizes()
# TestPixelSpecApply.test_spectral_function_validation()
# TestPixelSpecApply.test_single_pixel_image()


# %% test functions : pixel_array_apply


class TestPixelArrayApply:
    """Test class for pixel_array_apply function"""

    @staticmethod
    def simple_spectral_function(spectra_2d: np.ndarray) -> np.ndarray:
        """Simple test function that doubles the values"""
        return spectra_2d * 2

    @staticmethod
    def band_reduction_function(spectra_2d: np.ndarray) -> np.ndarray:
        """Test function that reduces bands (4 -> 2)"""
        return spectra_2d[:, :2]  # Keep only first 2 bands

    @staticmethod
    def band_expansion_function(spectra_2d: np.ndarray) -> np.ndarray:
        """Test function that expands bands (4 -> 6)"""
        result: np.ndarray = np.column_stack([spectra_2d, spectra_2d[:, :2]])
        return result  # Add first two bands as fifth and sixth

    @staticmethod
    def test_basic_functionality() -> None:
        """Test basic functionality with simple spectral function"""
        with tempfile.TemporaryDirectory() as temp_dir:
            # Create test files
            input_path: str = os.path.join(temp_dir, "input.tif")
            output_path: str = os.path.join(temp_dir, "output.tif")

            create_test_raster(input_path, width=8, height=8, bands=4)

            # Apply function
            pixel_array_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelArrayApply.simple_spectral_function,
                dtype="float32",
                tile_size=4,
                progress=False,
            )

            # Verify output exists
            assert os.path.exists(output_path)

            # Verify output content
            with rasterio.open(output_path) as dst:
                assert dst.count == 4  # Same number of bands
                assert dst.meta["dtype"] == "float32"

                # Read and verify data was doubled
                output_data: np.ndarray = dst.read()
                with rasterio.open(input_path) as src:
                    input_data: np.ndarray = src.read()

                np.testing.assert_allclose(output_data, input_data * 2, rtol=1e-6)

    @staticmethod
    def test_band_reduction() -> None:
        """Test function that reduces number of bands"""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path: str = os.path.join(temp_dir, "input.tif")
            output_path: str = os.path.join(temp_dir, "output.tif")

            create_test_raster(input_path, width=6, height=6, bands=4)

            pixel_array_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelArrayApply.band_reduction_function,
                tile_size=3,
                progress=False,
            )

            with rasterio.open(output_path) as dst:
                assert dst.count == 2  # Reduced from 4 to 2 bands

    @staticmethod
    def test_band_expansion() -> None:
        """Test function that increases number of bands"""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path: str = os.path.join(temp_dir, "input.tif")
            output_path: str = os.path.join(temp_dir, "output.tif")

            create_test_raster(input_path, width=5, height=5, bands=4)

            pixel_array_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelArrayApply.band_expansion_function,
                tile_size=2,
                progress=False,
            )

            with rasterio.open(output_path) as dst:
                assert dst.count == 6  # Expanded from 4 to 6 bands

    @staticmethod
    def test_different_dtypes() -> None:
        """Test with different output dtypes"""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path: str = os.path.join(temp_dir, "input.tif")
            output_path: str = os.path.join(temp_dir, "output.tif")

            create_test_raster(input_path, width=4, height=4, bands=4)

            pixel_array_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelArrayApply.simple_spectral_function,
                dtype="uint16",
                tile_size=2,
                progress=False,
            )

            with rasterio.open(output_path) as dst:
                assert dst.meta["dtype"] == "uint16"

    @staticmethod
    def test_default_tile_size() -> None:
        """Test with default tile size (-1)"""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path: str = os.path.join(temp_dir, "input.tif")
            output_path: str = os.path.join(temp_dir, "output.tif")

            create_test_raster(input_path, width=10, height=10, bands=4)

            pixel_array_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelArrayApply.simple_spectral_function,
                progress=False,
            )

            assert os.path.exists(output_path)

    @staticmethod
    def test_small_tile_size() -> None:
        """Test with small tile size"""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path: str = os.path.join(temp_dir, "input.tif")
            output_path: str = os.path.join(temp_dir, "output.tif")

            create_test_raster(input_path, width=7, height=7, bands=4)

            pixel_array_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelArrayApply.simple_spectral_function,
                tile_size=1,
                progress=False,
            )

            assert os.path.exists(output_path)

    @staticmethod
    def test_large_tile_size() -> None:
        """Test with tile size larger than image"""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path: str = os.path.join(temp_dir, "input.tif")
            output_path: str = os.path.join(temp_dir, "output.tif")

            create_test_raster(input_path, width=5, height=5, bands=4)

            pixel_array_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelArrayApply.simple_spectral_function,
                tile_size=10,
                progress=False,
            )

            assert os.path.exists(output_path)

    @staticmethod
    def test_invalid_tile_size() -> None:
        """Test with invalid tile size"""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path: str = os.path.join(temp_dir, "input.tif")
            output_path: str = os.path.join(temp_dir, "output.tif")

            create_test_raster(input_path, width=5, height=5, bands=4)

            with pytest.raises(ValueError, match="tile_size must be positive"):
                pixel_array_apply(
                    image_path=input_path,
                    output_path=output_path,
                    spectral_function=TestPixelArrayApply.simple_spectral_function,
                    tile_size=0,
                    progress=False,
                )

    @staticmethod
    def test_output_file_overwrite() -> None:
        """Test that existing output file is overwritten"""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path: str = os.path.join(temp_dir, "input.tif")
            output_path: str = os.path.join(temp_dir, "output.tif")

            create_test_raster(input_path, width=4, height=4, bands=4)

            # Create a dummy output file first
            with open(output_path, "w") as f:
                f.write("dummy content")

            pixel_array_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelArrayApply.simple_spectral_function,
                tile_size=2,
                progress=False,
            )

            # Should be a valid raster file, not text
            with rasterio.open(output_path) as dst:
                assert dst.count == 4

    @staticmethod
    def test_complex_spectral_function() -> None:
        """Test with a more complex spectral function"""

        def complex_function(spectra_2d: np.ndarray) -> np.ndarray:
            """Normalize each spectrum and compute mean"""
            normalized: np.ndarray = spectra_2d / np.linalg.norm(spectra_2d, axis=1, keepdims=True)
            mean_band: np.ndarray = np.mean(normalized, axis=1, keepdims=True)
            result: np.ndarray = np.column_stack([normalized, mean_band])
            return result

        with tempfile.TemporaryDirectory() as temp_dir:
            input_path: str = os.path.join(temp_dir, "input.tif")
            output_path: str = os.path.join(temp_dir, "output.tif")

            create_test_raster(input_path, width=4, height=4, bands=4)

            pixel_array_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=complex_function,
                tile_size=2,
                progress=False,
            )

            with rasterio.open(output_path) as dst:
                assert dst.count == 5  # 4 normalized bands + 1 mean band

    @staticmethod
    def test_single_pixel_image() -> None:
        """Test with a single pixel image"""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path: str = os.path.join(temp_dir, "input.tif")
            output_path: str = os.path.join(temp_dir, "output.tif")

            create_test_raster(input_path, width=1, height=1, bands=4)

            pixel_array_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelArrayApply.simple_spectral_function,
                tile_size=1,
                progress=False,
            )

            with rasterio.open(output_path) as dst:
                assert dst.count == 4
                assert dst.width == 1
                assert dst.height == 1

    @staticmethod
    def test_function_signature_validation() -> None:
        """Test that the function validates spectral_function signature"""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path: str = os.path.join(temp_dir, "input.tif")
            output_path: str = os.path.join(temp_dir, "output.tif")

            create_test_raster(input_path, width=3, height=3, bands=4)

            # Function with wrong signature (no parameters)
            def invalid_function() -> np.ndarray:
                result: np.ndarray = np.array([[1, 2]])
                return result

            # This should raise an error when the function is called
            with pytest.raises((TypeError, ValueError)):
                pixel_array_apply(
                    image_path=input_path,
                    output_path=output_path,
                    spectral_function=invalid_function,
                    tile_size=2,
                    progress=False,
                )


# %% Test - pixel_array_apply

# TestPixelArrayApply.test_basic_functionality()

# TestPixelArrayApply.test_band_reduction()
# TestPixelArrayApply.test_band_expansion()
# TestPixelArrayApply.test_complex_spectral_function()

# TestPixelArrayApply.test_different_dtypes()

# TestPixelArrayApply.test_default_tile_size()
# TestPixelArrayApply.test_small_tile_size()
# TestPixelArrayApply.test_large_tile_size()
# TestPixelArrayApply.test_invalid_tile_size()

# TestPixelArrayApply.test_single_pixel_image()

# TestPixelArrayApply.test_output_file_overwrite()
# TestPixelArrayApply.test_function_signature_validation()


# %% test functions : pixel_tensor_apply


@pytest.mark.skipif(not HAS_CUDA, reason="No GPU available")
class TestPixelTensorApply(unittest.TestCase):
    @staticmethod
    def simple_spectral_function(tensor: torch.Tensor) -> torch.Tensor:
        """Simple test function that doubles the values along axis 0."""
        return tensor * 2

    @staticmethod
    def band_reduction_function(tensor: torch.Tensor) -> torch.Tensor:
        """Test function that reduces bands by taking mean of every 2 bands."""
        result = []
        for i in range(0, tensor.shape[0], 2):
            if i + 1 < tensor.shape[0]:
                result.append((tensor[i] + tensor[i + 1]) / 2)
            else:
                result.append(tensor[i])
        return torch.stack(result)

    @staticmethod
    def identity_function(tensor: torch.Tensor) -> torch.Tensor:
        """Identity function for testing."""
        return tensor.clone()

    @staticmethod
    def spectral_normalization_function(tensor: torch.Tensor) -> torch.Tensor:
        """Test function that normalizes each band to [0, 1] range."""
        result = []
        for i in range(tensor.shape[0]):
            band = tensor[i]
            min_val = band.min()
            max_val = band.max()
            if max_val > min_val:
                normalized = (band - min_val) / (max_val - min_val)
            else:
                normalized = band
            result.append(normalized)
        return torch.stack(result)

    @staticmethod
    def test_basic_functionality() -> None:
        """Test basic functionality with a simple spectral function."""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "test_input.tif")
            output_path = os.path.join(temp_dir, "test_output.tif")

            create_test_raster(input_path, width=10, height=10, bands=3)

            pixel_tensor_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelTensorApply.simple_spectral_function,
                dtype="float32",
                tile_size=5,
                device="cuda",
                progress=False,
            )

            assert os.path.exists(output_path)

            with rasterio.open(output_path) as src:
                output_data = src.read()
                with rasterio.open(input_path) as src_input:
                    input_data = src_input.read().astype(np.float32)

                np.testing.assert_allclose(output_data, input_data * 2, rtol=1e-6)

    @staticmethod
    def test_band_reduction() -> None:
        """Test with a function that changes the number of bands."""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "test_input.tif")
            output_path = os.path.join(temp_dir, "test_output.tif")

            create_test_raster(input_path, width=8, height=8, bands=4)

            pixel_tensor_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelTensorApply.band_reduction_function,
                dtype="float32",
                tile_size=4,
                device="cuda",
                progress=False,
            )

            assert os.path.exists(output_path)

            with rasterio.open(output_path) as src:
                output_data = src.read()
                assert output_data.shape[0] == 2

    @staticmethod
    def test_identity() -> None:
        """Test identity function preserves input exactly."""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "test_input.tif")
            output_path = os.path.join(temp_dir, "test_output.tif")

            create_test_raster(input_path, width=6, height=6, bands=2)

            pixel_tensor_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelTensorApply.identity_function,
                dtype="uint16",
                tile_size=3,
                device="cuda",
                progress=False,
            )

            assert os.path.exists(output_path)

            with rasterio.open(input_path) as src_input:
                with rasterio.open(output_path) as src_output:
                    input_data = src_input.read()
                    output_data = src_output.read()

                    np.testing.assert_array_equal(input_data, output_data)

    @staticmethod
    def test_different_dtypes() -> None:
        """Test with different output data types."""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "test_input.tif")
            output_path_float32 = os.path.join(temp_dir, "test_output_float32.tif")
            output_path_float64 = os.path.join(temp_dir, "test_output_float64.tif")

            create_test_raster(input_path, width=5, height=5, bands=2)

            pixel_tensor_apply(
                image_path=input_path,
                output_path=output_path_float32,
                spectral_function=TestPixelTensorApply.simple_spectral_function,
                dtype="float32",
                tile_size=3,
                device="cuda",
                progress=False,
            )

            pixel_tensor_apply(
                image_path=input_path,
                output_path=output_path_float64,
                spectral_function=TestPixelTensorApply.simple_spectral_function,
                dtype="float64",
                tile_size=3,
                device="cuda",
                progress=False,
            )

            with rasterio.open(output_path_float32) as src:
                assert src.meta["dtype"] == "float32"

            with rasterio.open(output_path_float64) as src:
                assert src.meta["dtype"] == "float64"

    @staticmethod
    def test_different_tile_sizes() -> None:
        """Test with different tile sizes."""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "test_input.tif")
            output_path1 = os.path.join(temp_dir, "test_output1.tif")
            output_path2 = os.path.join(temp_dir, "test_output2.tif")

            create_test_raster(input_path, width=12, height=12, bands=3)

            pixel_tensor_apply(
                image_path=input_path,
                output_path=output_path1,
                spectral_function=TestPixelTensorApply.simple_spectral_function,
                dtype="float32",
                tile_size=4,
                device="cuda",
                progress=False,
            )

            pixel_tensor_apply(
                image_path=input_path,
                output_path=output_path2,
                spectral_function=TestPixelTensorApply.simple_spectral_function,
                dtype="float32",
                tile_size=6,
                device="cuda",
                progress=False,
            )

            with rasterio.open(output_path1) as src1:
                with rasterio.open(output_path2) as src2:
                    data1 = src1.read()
                    data2 = src2.read()
                    np.testing.assert_allclose(data1, data2, rtol=1e-6)

    @staticmethod
    def test_invalid_tile_size() -> None:
        """Test that invalid tile size raises ValueError."""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "test_input.tif")
            output_path = os.path.join(temp_dir, "test_output.tif")

            create_test_raster(input_path, width=10, height=10, bands=2)

            try:
                pixel_tensor_apply(
                    image_path=input_path,
                    output_path=output_path,
                    spectral_function=TestPixelTensorApply.identity_function,
                    dtype="float32",
                    tile_size=0,
                    device="cuda",
                    progress=False,
                )
                raise AssertionError("Expected ValueError")
            except ValueError:
                pass

    @staticmethod
    def test_output_overwrite() -> None:
        """Test that existing output file is overwritten."""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "test_input.tif")
            output_path = os.path.join(temp_dir, "test_output.tif")

            create_test_raster(input_path, width=5, height=5, bands=2)

            with open(output_path, "w") as f:
                f.write("dummy content")

            pixel_tensor_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelTensorApply.identity_function,
                dtype="float32",
                tile_size=3,
                device="cuda",
                progress=False,
            )

            with rasterio.open(output_path) as src:
                assert src.count == 2

    @staticmethod
    def test_edge_case_small_image() -> None:
        """Test with very small image size."""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "test_input.tif")
            output_path = os.path.join(temp_dir, "test_output.tif")

            create_test_raster(input_path, width=2, height=2, bands=1)

            pixel_tensor_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelTensorApply.simple_spectral_function,
                dtype="float32",
                tile_size=64,
                device="cuda",
                progress=False,
            )

            assert os.path.exists(output_path)

            with rasterio.open(input_path) as src_input:
                with rasterio.open(output_path) as src_output:
                    input_data = src_input.read().astype(np.float32)
                    output_data = src_output.read()

                    np.testing.assert_allclose(output_data, input_data * 2, rtol=1e-6)

    @staticmethod
    def test_spectral_normalization() -> None:
        """Test with a spectral normalization function."""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "test_input.tif")
            output_path = os.path.join(temp_dir, "test_output.tif")

            create_test_raster(input_path, width=8, height=8, bands=3)

            pixel_tensor_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelTensorApply.spectral_normalization_function,
                dtype="float32",
                tile_size=4,
                device="cuda",
                progress=False,
            )

            assert os.path.exists(output_path)

            with rasterio.open(output_path) as src:
                output_data = src.read()
                assert output_data.min() >= 0.0
                assert output_data.max() <= 1.0 + 1e-6

    @staticmethod
    def test_default_tile_size() -> None:
        """Test with default tile size (-1)."""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "test_input.tif")
            output_path = os.path.join(temp_dir, "test_output.tif")

            create_test_raster(input_path, width=10, height=10, bands=2)

            pixel_tensor_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelTensorApply.identity_function,
                dtype="float32",
                tile_size=-1,
                device="cuda",
                progress=False,
            )

            assert os.path.exists(output_path)

            with rasterio.open(input_path) as src_input:
                with rasterio.open(output_path) as src_output:
                    input_data = src_input.read().astype(np.float32)
                    output_data = src_output.read()

                    np.testing.assert_allclose(input_data, output_data, rtol=1e-6)

    @staticmethod
    def test_cpu_device() -> None:
        """Test with CPU device explicitly specified."""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "test_input.tif")
            output_path = os.path.join(temp_dir, "test_output.tif")

            create_test_raster(input_path, width=5, height=5, bands=2)

            pixel_tensor_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelTensorApply.simple_spectral_function,
                dtype="float32",
                tile_size=3,
                device="cpu",
                progress=False,
            )

            assert os.path.exists(output_path)

            with rasterio.open(input_path) as src_input:
                with rasterio.open(output_path) as src_output:
                    input_data = src_input.read().astype(np.float32)
                    output_data = src_output.read()

                    np.testing.assert_allclose(output_data, input_data * 2, rtol=1e-6)

    @staticmethod
    def test_large_tile_size() -> None:
        """Test with tile size larger than image dimensions."""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "test_input.tif")
            output_path = os.path.join(temp_dir, "test_output.tif")

            create_test_raster(input_path, width=3, height=3, bands=2)

            pixel_tensor_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelTensorApply.simple_spectral_function,
                dtype="float32",
                tile_size=10,
                device="cuda",
                progress=False,
            )

            assert os.path.exists(output_path)

            with rasterio.open(input_path) as src_input:
                with rasterio.open(output_path) as src_output:
                    input_data = src_input.read().astype(np.float32)
                    output_data = src_output.read()

                    np.testing.assert_allclose(output_data, input_data * 2, rtol=1e-6)


# %% Test - pixel_tensor_apply

# TestPixelTensorApply.test_basic_functionality()

# TestPixelTensorApply.test_band_reduction()
# TestPixelTensorApply.test_identity()
# TestPixelTensorApply.test_spectral_normalization()

# TestPixelTensorApply.test_different_dtypes()

# TestPixelTensorApply.test_default_tile_size()
# TestPixelTensorApply.test_different_tile_sizes()
# TestPixelTensorApply.test_invalid_tile_size()
# TestPixelTensorApply.test_large_tile_size()

# TestPixelTensorApply.test_edge_case_small_image()

# TestPixelTensorApply.test_cpu_device()

# TestPixelTensorApply.test_output_overwrite()


# %% test functions : pixel_tensor_hyper_apply


@pytest.mark.skipif(not HAS_CUDA, reason="No GPU available")
class TestPixelTensorHyperApply(unittest.TestCase):
    @staticmethod
    def simple_spectral_function(tensor: torch.Tensor) -> torch.Tensor:
        """Simple spectral function for normalization and scaling"""
        # Simple function: normalize and scale
        tensor = tensor.float()
        mean = tensor.mean(dim=1, keepdim=True)
        std = tensor.std(dim=1, keepdim=True)
        return ((tensor - mean) / (std + 1e-8)) * 100

    @staticmethod
    def band_reduction_function(tensor: torch.Tensor) -> torch.Tensor:
        """Function that reduces bands"""
        tensor = tensor.float()
        # Simple mean/std normalization
        normalized = (tensor - tensor.mean(dim=1, keepdim=True)) / (tensor.std(dim=1, keepdim=True) + 1e-8)
        # Return first 2 "components"
        return normalized[:, :2, :]

    @staticmethod
    def identity_function(tensor: torch.Tensor) -> torch.Tensor:
        return tensor.clone()

    @staticmethod
    def test_basic_functionality_cpu() -> None:
        """Test basic functionality with identity function on CPU."""
        with tempfile.TemporaryDirectory() as temp_dir:
            # Create test raster
            input_path = os.path.join(temp_dir, "test_input.tif")
            output_path = os.path.join(temp_dir, "test_output.tif")
            create_test_raster(input_path, width=10, height=5, bands=3, dtype="uint16")
            pixel_tensor_hyper_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelTensorHyperApply.identity_function,
                dtype="float32",
                tile_size=2,
                device="cpu",
                progress=False,
            )

            # Verify output exists and has correct properties
            assert os.path.exists(output_path)

            with rasterio.open(input_path) as src, rasterio.open(output_path) as dst:
                assert dst.count == src.count  # Same number of bands
                assert dst.width == src.width
                assert dst.height == src.height
                assert dst.meta["dtype"] == "float32"

    @staticmethod
    def test_basic_functionality_gpu() -> None:
        """Test basic functionality with identity function on GPU."""
        with tempfile.TemporaryDirectory() as temp_dir:
            # Create test raster
            input_path = os.path.join(temp_dir, "test_input.tif")
            output_path = os.path.join(temp_dir, "test_output.tif")
            create_test_raster(input_path, width=10, height=5, bands=3, dtype="uint16")
            pixel_tensor_hyper_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelTensorHyperApply.identity_function,
                dtype="float32",
                tile_size=2,
                device="cuda",
                progress=False,
            )

            # Verify output exists and has correct properties
            assert os.path.exists(output_path)

            with rasterio.open(input_path) as src, rasterio.open(output_path) as dst:
                assert dst.count == src.count  # Same number of bands
                assert dst.width == src.width
                assert dst.height == src.height
                assert dst.meta["dtype"] == "float32"

    @staticmethod
    def test_band_reduction_function_gpu() -> None:
        """Test with a function that reduces number of bands on GPU."""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "test_input.tif")
            output_path = os.path.join(temp_dir, "test_output.tif")
            create_test_raster(input_path, width=8, height=4, bands=5, dtype="uint16")
            pixel_tensor_hyper_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelTensorHyperApply.band_reduction_function,
                dtype="float32",
                tile_size=3,
                device="cuda",
                progress=False,
            )

            with rasterio.open(output_path) as dst:
                assert dst.count == 2  # Reduced from 5 to 2 bands

    @staticmethod
    def test_different_dtypes_gpu() -> None:
        """Test with different output data types on GPU."""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "test_input.tif")
            # Test float32
            output_path_float32 = os.path.join(temp_dir, "test_output_float32.tif")
            create_test_raster(input_path, width=6, height=3, bands=2, dtype="uint16")
            pixel_tensor_hyper_apply(
                image_path=input_path,
                output_path=output_path_float32,
                spectral_function=TestPixelTensorHyperApply.simple_spectral_function,
                dtype="float32",
                tile_size=1,
                device="cuda",
                progress=False,
            )
            with rasterio.open(output_path_float32) as dst:
                assert dst.meta["dtype"] == "float32"

            # Test float64
            output_path_float64 = os.path.join(temp_dir, "test_output_float64.tif")
            create_test_raster(input_path, width=6, height=3, bands=2, dtype="uint16")
            pixel_tensor_hyper_apply(
                image_path=input_path,
                output_path=output_path_float64,
                spectral_function=TestPixelTensorHyperApply.simple_spectral_function,
                dtype="float64",
                tile_size=1,
                device="cuda",
                progress=False,
            )
            with rasterio.open(output_path_float64) as dst:
                assert dst.meta["dtype"] == "float64"

    @staticmethod
    def test_different_tile_sizes_gpu() -> None:
        """Test with different tile sizes on GPU."""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "test_input.tif")

            # Test tile_size = 1
            output_path_1 = os.path.join(temp_dir, "test_output_1.tif")
            create_test_raster(input_path, width=10, height=10, bands=3, dtype="uint16")
            pixel_tensor_hyper_apply(
                image_path=input_path,
                output_path=output_path_1,
                spectral_function=TestPixelTensorHyperApply.identity_function,
                dtype="float32",
                tile_size=1,
                device="cuda",
                progress=False,
            )
            assert os.path.exists(output_path_1)

            # Test tile_size = 2
            output_path_2 = os.path.join(temp_dir, "test_output_2.tif")
            create_test_raster(input_path, width=10, height=10, bands=3, dtype="uint16")
            pixel_tensor_hyper_apply(
                image_path=input_path,
                output_path=output_path_2,
                spectral_function=TestPixelTensorHyperApply.identity_function,
                dtype="float32",
                tile_size=2,
                device="cuda",
                progress=False,
            )
            assert os.path.exists(output_path_2)

            # Test tile_size = 5
            output_path_5 = os.path.join(temp_dir, "test_output_5.tif")
            create_test_raster(input_path, width=10, height=10, bands=3, dtype="uint16")
            pixel_tensor_hyper_apply(
                image_path=input_path,
                output_path=output_path_5,
                spectral_function=TestPixelTensorHyperApply.identity_function,
                dtype="float32",
                tile_size=5,
                device="cuda",
                progress=False,
            )
            assert os.path.exists(output_path_5)

            # Test tile_size = 10
            output_path_10 = os.path.join(temp_dir, "test_output_10.tif")
            create_test_raster(input_path, width=10, height=10, bands=3, dtype="uint16")
            pixel_tensor_hyper_apply(
                image_path=input_path,
                output_path=output_path_10,
                spectral_function=TestPixelTensorHyperApply.identity_function,
                dtype="float32",
                tile_size=10,
                device="cuda",
                progress=False,
            )
            assert os.path.exists(output_path_10)

    @staticmethod
    def test_default_tile_size_gpu() -> None:
        """Test with default tile size (-1) on GPU."""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "test_input.tif")
            output_path = os.path.join(temp_dir, "test_output.tif")
            create_test_raster(input_path, width=5, height=5, bands=2, dtype="uint16")
            pixel_tensor_hyper_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelTensorHyperApply.identity_function,
                dtype="float32",
                tile_size=-1,  # Default
                device="cuda",
                progress=False,
            )

            assert os.path.exists(output_path)

    @staticmethod
    def test_invalid_tile_size_gpu() -> None:
        """Test that invalid tile size raises ValueError on GPU."""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "test_input.tif")
            output_path = os.path.join(temp_dir, "test_output.tif")
            create_test_raster(input_path, width=5, height=5, bands=2, dtype="uint16")
            with pytest.raises(ValueError):
                pixel_tensor_hyper_apply(
                    image_path=input_path,
                    output_path=output_path,
                    spectral_function=TestPixelTensorHyperApply.identity_function,
                    dtype="float32",
                    tile_size=0,  # Invalid
                    device="cuda",
                    progress=False,
                )

    @staticmethod
    def test_output_overwrite_gpu() -> None:
        """Test that existing output file is overwritten on GPU."""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "test_input.tif")
            output_path = os.path.join(temp_dir, "test_output.tif")
            create_test_raster(input_path, width=3, height=2, bands=1, dtype="uint16")

            # Create a dummy output file first
            with open(output_path, "w") as f:
                f.write("dummy content")

            pixel_tensor_hyper_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelTensorHyperApply.identity_function,
                dtype="float32",
                tile_size=1,
                device="cuda",
                progress=False,
            )

            # Verify the file was overwritten with valid raster data
            assert os.path.exists(output_path)
            with rasterio.open(output_path) as dst:
                assert dst.count == 1

    @staticmethod
    def test_small_raster_edge_case_gpu() -> None:
        """Test with very small raster dimensions on GPU."""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "test_input.tif")
            output_path = os.path.join(temp_dir, "test_output.tif")

            # Create a very small raster
            create_test_raster(input_path, width=1, height=1, bands=1, dtype="uint16")
            pixel_tensor_hyper_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelTensorHyperApply.identity_function,
                dtype="float32",
                tile_size=1,
                device="cuda",
                progress=False,
            )

            # Verify output
            with rasterio.open(output_path) as dst:
                assert dst.width == 1
                assert dst.height == 1
                assert dst.count == 1

    @staticmethod
    def test_large_tile_size_gpu() -> None:
        """Test with tile size larger than image height on GPU."""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "test_input.tif")
            output_path = os.path.join(temp_dir, "test_output.tif")

            create_test_raster(input_path, width=3, height=2, bands=2, dtype="uint16")

            pixel_tensor_hyper_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelTensorHyperApply.identity_function,
                dtype="float32",
                tile_size=10,  # Larger than image height
                device="cuda",
                progress=False,
            )

            assert os.path.exists(output_path)

    @staticmethod
    def test_complex_spectral_function_gpu() -> None:
        """Test with a more complex spectral function on GPU."""
        with tempfile.TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, "test_input.tif")
            output_path = os.path.join(temp_dir, "test_output.tif")

            create_test_raster(input_path, width=8, height=6, bands=4, dtype="uint16")

            pixel_tensor_hyper_apply(
                image_path=input_path,
                output_path=output_path,
                spectral_function=TestPixelTensorHyperApply.simple_spectral_function,
                dtype="float32",
                tile_size=3,
                device="cuda",
                progress=False,
            )

            # Verify output has correct shape and properties
            with rasterio.open(output_path) as dst:
                assert dst.count == 4  # Same number of bands
                assert dst.width == 8
                assert dst.height == 6
                assert dst.meta["dtype"] == "float32"


# %% Test - pixel_tensor_hyper_apply

# TestPixelTensorHyperApply.test_basic_functionality_cpu()
# TestPixelTensorHyperApply.test_basic_functionality_gpu()
# TestPixelTensorHyperApply.test_band_reduction_function_gpu()
# TestPixelTensorHyperApply.test_complex_spectral_function_gpu()

# TestPixelTensorHyperApply.test_different_dtypes_gpu()

# TestPixelTensorHyperApply.test_different_tile_sizes_gpu()
# TestPixelTensorHyperApply.test_default_tile_size_gpu()
# TestPixelTensorHyperApply.test_invalid_tile_size_gpu()
# TestPixelTensorHyperApply.test_large_tile_size_gpu()

# TestPixelTensorHyperApply.test_output_overwrite_gpu()

# TestPixelTensorHyperApply.test_small_raster_edge_case_gpu()


# %% test functions : pixel_apply


class TestPixelApply:
    @staticmethod
    def dummy_spectral_function_spec(spectrum: np.ndarray) -> np.ndarray:
        """Dummy spectral function for 'spec' type."""
        return spectrum * 2

    @staticmethod
    def dummy_spectral_function_array(spectra: np.ndarray) -> np.ndarray:
        """Dummy spectral function for 'array' type."""
        return spectra * 2

    @staticmethod
    def dummy_spectral_function_tensor(tensor: torch.Tensor) -> torch.Tensor:
        """Dummy spectral function for 'tensor' type."""
        return tensor * 2

    @staticmethod
    def test_invalid_input_path() -> None:
        """Test handling of non-existent input file."""
        with pytest.raises(ValueError, match="Image file does not exist"):
            pixel_apply(
                image_path="non_existent.tif",
                spectral_function=TestPixelApply.dummy_spectral_function_spec,
                function_type="spec",
            )

    @staticmethod
    def test_invalid_output_dir() -> None:
        """Test handling of non-existent output directory."""
        # Create a temporary input file
        with tempfile.NamedTemporaryFile(suffix=".tif", delete=False) as tmp:
            create_test_raster(tmp.name, 10, 10, 10)
            input_path = tmp.name

        try:
            with pytest.raises(ValueError, match="Directory does not exist"):
                pixel_apply(
                    image_path=input_path,
                    spectral_function=TestPixelApply.dummy_spectral_function_spec,
                    function_type="spec",
                    output_path="/invalid/path/output.tif",
                )
        finally:
            os.remove(input_path)

    @staticmethod
    def test_invalid_tile_size() -> None:
        """Test handling of invalid tile size."""
        # Create temporary input and output paths
        with tempfile.NamedTemporaryFile(suffix=".tif", delete=False) as tmp_input:
            create_test_raster(tmp_input.name, 10, 10, 10)
            input_path = tmp_input.name

        with tempfile.TemporaryDirectory() as tmpdir:
            output_path = os.path.join(tmpdir, "output.tif")

            try:
                with pytest.raises(ValueError, match="tile_size must be positive integer"):
                    pixel_apply(
                        image_path=input_path,
                        spectral_function=TestPixelApply.dummy_spectral_function_spec,
                        function_type="spec",
                        output_path=output_path,
                        tile_size=0,
                    )
            finally:
                os.remove(input_path)

    @staticmethod
    def test_default_output_path() -> None:
        """Test generation of default output path."""
        # Create a temporary input file
        with tempfile.NamedTemporaryFile(suffix=".tif", delete=False) as tmp:
            create_test_raster(tmp.name, 10, 10, 10)
            input_path = tmp.name

        try:
            result = pixel_apply(
                image_path=input_path,
                spectral_function=TestPixelApply.dummy_spectral_function_spec,
                function_type="spec",
                return_output_path=True,
                progress=False,
            )
            assert result.endswith("_px_app_dummy_spectral_function_spec.tif")
            assert os.path.exists(result)

        finally:
            os.remove(input_path)
            if result is not None:
                os.remove(result)

    @staticmethod
    def test_spec_function_type() -> None:
        """Test processing with 'spec' function type."""
        # Create temporary input and output paths
        with tempfile.NamedTemporaryFile(suffix=".tif", delete=False) as tmp_input:
            create_test_raster(tmp_input.name, 10, 10, 10)
            input_path = tmp_input.name

        with tempfile.TemporaryDirectory() as tmpdir:
            output_path = os.path.join(tmpdir, "output.tif")

            try:
                result = pixel_apply(
                    image_path=input_path,
                    spectral_function=TestPixelApply.dummy_spectral_function_spec,
                    function_type="spec",
                    output_path=output_path,
                    progress=False,
                )
                assert result == output_path

            finally:
                os.remove(input_path)
                if result is not None:
                    os.remove(result)

    @staticmethod
    def test_array_function_type() -> None:
        """Test processing with 'array' function type."""
        # Create temporary input and output paths
        with tempfile.NamedTemporaryFile(suffix=".tif", delete=False) as tmp_input:
            create_test_raster(tmp_input.name, 10, 10, 10)
            input_path = tmp_input.name

        with tempfile.TemporaryDirectory() as tmpdir:
            output_path = os.path.join(tmpdir, "output.tif")

            try:
                result = pixel_apply(
                    image_path=input_path,
                    spectral_function=TestPixelApply.dummy_spectral_function_array,
                    function_type="array",
                    output_path=output_path,
                    progress=False,
                )
                assert result == output_path

            finally:
                os.remove(input_path)

    @staticmethod
    @pytest.mark.skipif(not HAS_CUDA, reason="No GPU available")
    def test_tensor_function_type() -> None:
        """Test processing with 'tensor' function type."""
        # Create temporary input and output paths
        with tempfile.NamedTemporaryFile(suffix=".tif", delete=False) as tmp_input:
            create_test_raster(tmp_input.name, 10, 10, 10)
            input_path = tmp_input.name

        with tempfile.TemporaryDirectory() as tmpdir:
            output_path = os.path.join(tmpdir, "output.tif")

            try:
                result = pixel_apply(
                    image_path=input_path,
                    spectral_function=TestPixelApply.dummy_spectral_function_tensor,
                    function_type="tensor",
                    output_path=output_path,
                    progress=False,
                )
                assert result == output_path

            finally:
                os.remove(input_path)

    @staticmethod
    @pytest.mark.skipif(not HAS_CUDA, reason="No GPU available")
    def test_tensor_hyper_function_type() -> None:
        """Test processing with 'tensor_hyper' function type."""
        # Create temporary input and output paths
        with tempfile.NamedTemporaryFile(suffix=".tif", delete=False) as tmp_input:
            create_test_raster(tmp_input.name, 10, 10, 10)
            input_path = tmp_input.name

        with tempfile.TemporaryDirectory() as tmpdir:
            output_path = os.path.join(tmpdir, "output.tif")

            try:
                result = pixel_apply(
                    image_path=input_path,
                    spectral_function=TestPixelApply.dummy_spectral_function_tensor,
                    function_type="tensor_hyper",
                    output_path=output_path,
                    progress=False,
                )
                assert result == output_path

            finally:
                os.remove(input_path)

    @staticmethod
    def test_invalid_function_type() -> None:
        """Test handling of invalid function type."""
        # Create temporary input and output paths
        with tempfile.NamedTemporaryFile(suffix=".tif", delete=False) as tmp_input:
            create_test_raster(tmp_input.name, 10, 10, 10)
            input_path = tmp_input.name

        with tempfile.TemporaryDirectory() as tmpdir:
            output_path = os.path.join(tmpdir, "output.tif")

            try:
                with pytest.raises(ValueError, match="Invalid function_type"):
                    pixel_apply(
                        image_path=input_path,
                        spectral_function=TestPixelApply.dummy_spectral_function_spec,
                        function_type="invalid",
                        output_path=output_path,
                    )
            finally:
                os.remove(input_path)

    @staticmethod
    def test_no_return_output_path() -> None:
        """Test behavior when return_output_path is False."""
        # Create temporary input and output paths
        with tempfile.NamedTemporaryFile(suffix=".tif", delete=False) as tmp_input:
            create_test_raster(tmp_input.name, 10, 10, 10)
            input_path = tmp_input.name

        with tempfile.TemporaryDirectory() as tmpdir:
            output_path = os.path.join(tmpdir, "output.tif")

            try:
                result = pixel_apply(
                    image_path=input_path,
                    spectral_function=TestPixelApply.dummy_spectral_function_spec,
                    function_type="spec",
                    output_path=output_path,
                    return_output_path=False,
                    progress=False,
                )
                assert result is None

            finally:
                os.remove(input_path)
                if result is not None:
                    os.remove(result)


# %% Test - pixel_apply

# TestPixelApply.test_default_output_path()
# TestPixelApply.test_invalid_input_path()
# TestPixelApply.test_invalid_output_dir()
# TestPixelApply.test_no_return_output_path()

# TestPixelApply.test_invalid_tile_size()

# TestPixelApply.test_spec_function_type()
# TestPixelApply.test_array_function_type()
# TestPixelApply.test_tensor_function_type()
# TestPixelApply.test_tensor_hyper_function_type()
# TestPixelApply.test_invalid_function_type()


# %% Test main

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
SpecPipe - A file-based pipeline for efficient batch processing and modeling of hyperspectral images
"""

# Package meta
__version__ = "0.1.0"
__author__ = "Siwei Luo"
__license__ = "MIT"

# Imports
__all__ = [
    ## Core pipeline tools
    ### Core class
    "SpecExp",
    "SpecPipe",
    "ModelEva",
    ### SpecIO functions
    "search_file",
    "envi_roi_coords",
    "shp_roi_coords",
    ### Raster operation
    "croproi",
    "pixel_apply",
    ## Spectral statistics tools
    ### ROI statistics
    "round_digit",
    "roispec",
    "pixcount",
    "nderiv",
    "moment2d",
    "bandhist",
    "Stats2d",
    "roi_mean",
    "roi_std",
    "roi_median",
    "spectral_angle",
    "arr_spectral_angles",
    ## Mock example data
    "create_example_raster",
    "create_example_roi_xml",
    "create_example_spec_exp",
]

# Components
## Core pipeline tools
## Example data generator
from .example_data import create_example_raster, create_example_roi_xml, create_example_spec_exp
from .modeleva import ModelEva
from .rasterop import croproi, pixel_apply

## Spectral statistics tools
from .roistats import (
    Stats2d,
    arr_spectral_angles,
    bandhist,
    moment2d,
    nderiv,
    pixcount,
    roi_mean,
    roi_median,
    roi_std,
    roispec,
    round_digit,
    spectral_angle,
)
from .specexp import SpecExp
from .specio import envi_roi_coords, search_file, shp_roi_coords
from .specpipe import SpecPipe

## Spectral processing tools



# -*- coding: utf-8 -*-
"""
Example data generator for SpecPipe

Copyright (c) 2025 Siwei Luo. MIT License.
"""

# OS Files
import os

# Typing
from typing import Optional, Union

# Testing third
import numpy as np
import pandas as pd

# Rasters
import rasterio

# Self
from .specexp import SpecExp
from .specio import roi_to_envi_xml, silent, simple_type_validator

# %% test helper functions


# Test helper functions : create_test_raster
@simple_type_validator
def create_test_raster(
    raster_path: str,
    width: int = 100,
    height: int = 50,
    bands: int = 4,
    incl_nodata: Union[None, int, float] = None,
    nodata_value: Union[int, float] = 0,
    dtype: Union[str, type] = "uint16",
    data: Optional[np.ndarray] = None,
) -> str:
    """Create a test raster file with by default 100x50 dimensions and 4 bands."""
    # Validate raster path
    raster_path = raster_path.replace("\\", "/").replace("//", "/")
    if (raster_path[-4:] != ".tif") and (raster_path[-5:] != ".tiff"):
        raise ValueError(f"raster_path must have .tif or .tiff extension, got: {raster_path}")
    if not os.path.exists(os.path.dirname(raster_path)):
        raise ValueError(f"raster path directory does not exist: {os.path.dirname(raster_path)}")

    # Validate raster data
    if data is None:
        # Mock image data
        np.random.seed(42)
        data = np.array(
            [
                [
                    (
                        np.array([1080 / bands * j + 360 / bands * int(i / 5)] * width)
                        + (np.random.rand(width) * 120 / bands)
                    )
                    .astype(dtype)
                    .tolist()
                    for i in range(height)
                ]
                for j in range(bands)
            ]
        )
    elif type(data) is np.ndarray:
        data = np.array(data).astype(dtype)
        if data.shape != (bands, height, width):
            raise ValueError(
                f"Given raster data shape {data.shape} does not match specified raster dimensions (bands={bands}, \
                    height={height}, width={width})."
            )
    else:
        raise ValueError(f"Given raster data must be numpy.ndarray, got type: {type(data)}.")

    # Set some nodata values
    if incl_nodata is not None:
        data[0, :, :] = incl_nodata
        if bands > 3:
            data[-1, :, :] = incl_nodata

    # Define transform (georeferencing)
    transform = rasterio.transform.from_bounds(0, 0, width, height, width=width, height=height)

    # Create metadata
    meta = {
        "driver": "GTiff",
        "dtype": dtype,
        "nodata": nodata_value,
        "width": width,
        "height": height,
        "count": bands,
        "crs": None,
        "transform": transform,
    }

    # Write to file
    with rasterio.open(raster_path, "w", **meta) as dst:
        dst.write(data)

    return raster_path


# Alias
create_example_raster = create_test_raster


# Test helper functions : create_test_roi_xml
@simple_type_validator
def create_test_roi_xml(
    xml_path: str,
    raster_width: int = 10,
    raster_height: int = 50,
    roi_count: int = 10,
    roi_list: Optional[list[dict]] = None,
    return_path: bool = True,
    return_roi_list: bool = False,
) -> Union[str, list[dict], tuple[str, list[dict]], None]:
    """Create a test ENVI ROI xml file with by default 10 ROIs."""
    # Create line of mock ROIs in the raster
    height_increment = int(raster_height / roi_count)
    if height_increment < 1:
        raise ValueError("raster_height / roi_count must be at least 1.")
    width_range = min(5, raster_width)

    if roi_list is None:
        roi_list = []
        # Test ROI data
        for i in range(roi_count):
            roi_dict = {
                "name": f"ROI_{i + 1}",
                "crs": "none",
                "color": None,
                "type": "polygon",
                "coordinates": [
                    [
                        (0.0, height_increment * i),
                        (width_range, height_increment * i),
                        (width_range, height_increment * (i + 1)),
                        (0.0, height_increment * (i + 1)),
                        (0.0, height_increment * i),
                    ]
                ],
            }
            roi_list.append(roi_dict)

    # Save ROI xml file
    path_out = roi_to_envi_xml(file_path=xml_path, roi_list=roi_list)
    path_out = str(path_out)

    # Output
    if return_path and not return_roi_list:
        return path_out
    elif not return_path and return_roi_list:
        return roi_list
    elif return_path and return_roi_list:
        return path_out, roi_list
    else:
        return None


# Alias
create_example_roi_xml = create_test_roi_xml


# Test helper functions : create_test_spec_exp
@silent
@simple_type_validator
def create_test_spec_exp(dir_path: str, sample_n: int = 10, n_bands: int = 4, is_regression: bool = True) -> SpecExp:
    """Create a standard test SpecExp instance."""
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)

    exp1: SpecExp = SpecExp(dir_path, log_loading=False)

    # # Add test
    # group
    exp1.add_groups(["test_group"])

    # Mock image data
    np.random.seed(42)
    mock_img_data = np.array(
        [
            [
                (np.array([1080 / n_bands * j + 360 / n_bands * i] * 10) + (np.random.rand(10) * 120 / n_bands))
                .astype(np.uint16)
                .tolist()
                for i in range(5 * sample_n)
            ]
            for j in range(n_bands)
        ]
    )

    for i in range(5 * sample_n):
        mock_img_data[-int(n_bands / 3) :, i, :] = mock_img_data[-int(n_bands / 3) :, i, :] * (i // 5 % 2 + 0.5)

    # Mock image
    img_path = dir_path + "/test_img.tif"
    create_test_raster(
        raster_path=img_path,
        width=mock_img_data.shape[2],
        height=mock_img_data.shape[1],
        bands=n_bands,
        data=mock_img_data,
    )
    exp1.add_images("test_group", ["test_img.tif"], dir_path)

    # ROIs
    roi_path = dir_path + "/test_roi.xml"
    create_test_roi_xml(roi_path, roi_count=sample_n)
    exp1.add_rois_by_file("test_group", [roi_path], "test_img.tif")

    # Samples
    dflb = exp1.ls_sample_labels()
    assert isinstance(dflb, pd.DataFrame)
    dflb.iloc[:, 1] = [f"sample_{str(i + 1)}" for i in range(len(dflb))]
    exp1.sample_labels = dflb  # type: ignore[assignment]
    # Auto-conversion in setter

    # Target values
    dft = exp1.ls_sample_targets()
    assert isinstance(dft, pd.DataFrame)
    if is_regression:
        dft["Target_value"] = list(range(len(dft)))
    else:
        dft["Target_value"] = [["a", "b"][int(i % 2)] for i in range(len(dft))]
    exp1.sample_targets_from_df(dft)

    return exp1


# Alias
create_example_spec_exp = create_test_spec_exp
